#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <png.h>

#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static void usage(void);
static int get_int_value(char *str, int *value);
#ifndef PICKUP_COLOR
static void create_palette(png_byte palette[256][3]);
static void change_color(png_byte palette[256][3], png_byte *bitmap, png_uint_32 width, png_uint_32 height);
static void change_color_and_alpha(png_byte palette[256][3], png_byte alpha_table[255], png_byte *bitmap, png_uint_32 width, png_uint_32 height);
#endif
static int load_png_file(const char *filename, png_byte **bitmap, png_uint_32 *width, png_uint_32 *height);
#ifndef PICKUP_COLOR
static int save_png_file(const char *filename, const png_byte *bitmap, png_uint_32 width, png_uint_32 height);
#endif


int main(int argc, char **argv)
{
#ifdef PICKUP_COLOR
	int num_cmd_params = 3;
#else
	int num_cmd_params = 7;
#endif
	png_byte *bitmap = NULL;
	png_uint_32 width;
	png_uint_32 height;
	const char *ref_file = NULL;
	int color_x = 0;
	int color_y = 0;
#ifndef PICKUP_COLOR
	int border_x = 0;
	int border_y = 0;
	const char *in_file = NULL;
	const char *out_file = NULL;
	png_byte palette[256][3];
	int alpha = 255;
#endif

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
#ifndef PICKUP_COLOR
			} else if (strcmp(*argv, "-a") == 0 || strcmp(*argv, "--alpha") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("ERROR: アルファ値倍率がありません\n");
					return 1;
				}
				alpha = strtol(*argv, &endptr, 0);
				if (endptr == NULL || *endptr != '\0' || alpha < 0) {
					printf("ERROR: 不正なアルファ値倍率です: %s\n", *argv);
					return 1;
				}
#endif
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
#ifdef PICKUP_COLOR
			if (num_cmd_params == 3) {
				ref_file = *argv;
			} else if (num_cmd_params == 2) {
				if (get_int_value(*argv, &color_x) < 0) {
					printf("ERRPR: 不正なX座標です: %s\n", *argv);
					return 1;
				}
			} else {
				if (get_int_value(*argv, &color_y) < 0) {
					printf("ERRPR: 不正なY座標です: %s\n", *argv);
					return 1;
				}
			}
#else
			if (num_cmd_params == 7) {
				ref_file = *argv;
			} else if (num_cmd_params == 6) {
				if (get_int_value(*argv, &color_x) < 0) {
					printf("ERRPR: 不正な文字色X座標です: %s\n", *argv);
					return 1;
				}
			} else if (num_cmd_params == 5) {
				if (get_int_value(*argv, &color_y) < 0) {
					printf("ERRPR: 不正な文字色Y座標です: %s\n", *argv);
					return 1;
				}
			} else if (num_cmd_params == 4) {
				if (get_int_value(*argv, &border_x) < 0) {
					printf("ERRPR: 不正な輪郭色X座標です: %s\n", *argv);
					return 1;
				}
			} else if (num_cmd_params == 3) {
				if (get_int_value(*argv, &border_y) < 0) {
					printf("ERRPR: 不正な輪郭色Y座標です: %s\n", *argv);
					return 1;
				}
			} else if (num_cmd_params == 2) {
				in_file = *argv;
			} else {
				out_file = *argv;
			}
#endif
			num_cmd_params--;
		}
	}

#ifdef PICKUP_COLOR
	if (ref_file == NULL)
#else
	if (ref_file == NULL || in_file == NULL || out_file == NULL)
#endif
	{
		usage();
		return 1;
	}

	if (load_png_file(ref_file, &bitmap, &width, &height) < 0) {
		printf("ERROR: 読み込みエラー: %s\n", ref_file);
		return 1;
	}

#ifdef PICKUP_COLOR
	if (width <= color_x || height <= color_y) {
		printf("ERRPR: 画像外の座標です\n");
		return 1;
	} else {
		const unsigned char *ptr = bitmap + (color_y * width + color_x) * 4;
		printf("(%d,%d) : RGBA %3.1d %3.1d %3.1d %3.1d\n", color_x, color_y, (int)ptr[0], (int)ptr[1], (int)ptr[2], (int)ptr[3]);
	}
#else
	if (width <= color_x || height <= color_y || width <= border_x || height <= border_y) {
		printf("ERRPR: 画像外の座標です\n");
		return 1;
	}

	memcpy(palette[255], bitmap + (color_y * width + color_x) * 4, 3);
	memcpy(palette[0], bitmap + (border_y * width + border_x) * 4, 3);

	create_palette(palette);
#endif

	free(bitmap);
	bitmap = NULL;

#ifndef PICKUP_COLOR
	if (load_png_file(in_file, &bitmap, &width, &height) < 0) {
		printf("ERROR: 読み込みエラー: %s\n", in_file);
		return 1;
	}

	if (alpha == 255) {
		change_color(palette, bitmap, width, height);
	} else {
		png_byte alpha_table[255];
		int i;

		for (i = 0; i <= 255; i++) {
			int value = i * alpha / 255;
			if (value < 0) {
				value = 0;
			} else if (255 < value) {
				value = 255;
			}
			alpha_table[i] = (png_byte)value;
		}

		change_color_and_alpha(palette, alpha_table, bitmap, width, height);
	}

	if (save_png_file(out_file, bitmap, width, height) < 0) {
		printf("ERROR: 書き込みエラー: %s\n", out_file);
		return 1;
	}
#endif

	return 0;
}

static void usage(void)
{
#ifdef PICKUP_COLOR
	printf("使用方法: pickup_color [オプション] 参照ファイル X座標 Y座標\n");
	printf("機    能: PNG画像の指定座標の色を表示します\n");
#else
	printf("使用方法: change_color [オプション] 参照ファイル 文字色X座標 文字色Y座標\n");
	printf("                              輪郭色X座標 輪郭色Y座標 入力ファイル 出力ファイル\n");
	printf("機    能: PNG画像の色を変更します\n");
#endif
	printf("オプション:\n");
#ifndef PICKUP_COLOR
	printf("    -a|--alpha 倍率     全体のアルファ値を×倍率÷255します\n");
	printf("                        倍率が255より小さいときは透明度が上がり\n");
	printf("                        倍率が255より大きいときは透明度が下がります\n");
#endif
	printf("    -?|--help           使用方法を表示して終了します\n");
}

static int get_int_value(char *str, int *value)
{
	char *endptr;

	*value = strtoul(str, &endptr, 0);
	if (endptr == NULL || *endptr != '\0') {
		return -1;
	}

	return 0;
}

#ifndef PICKUP_COLOR
static void create_palette(png_byte palette[256][3])
{
	int i;
	int j;

	for (i = 1; i <= 254; i++) {
		for (j = 0; j < 3; j++) {
			palette[i][j] = ((int)palette[0][j] * (255 - i) + (int)palette[255][j] * i) / 255;
		}
	}
}

static void change_color(png_byte palette[256][3], png_byte *bitmap, png_uint_32 width, png_uint_32 height)
{
	png_uint_32 w;
	png_uint_32 h;
	png_byte *ptr = bitmap;

	for (h = 0; h < height; h++) {
		for (w = 0; w < width; w++) {
			*ptr = palette[*ptr][0];
			ptr++;
			*ptr = palette[*ptr][1];
			ptr++;
			*ptr = palette[*ptr][2];
			ptr++;
			ptr++;
		}
	}
}

static void change_color_and_alpha(png_byte palette[256][3], png_byte alpha_table[255], png_byte *bitmap, png_uint_32 width, png_uint_32 height)
{
	png_uint_32 w;
	png_uint_32 h;
	png_byte *ptr = bitmap;

	for (h = 0; h < height; h++) {
		for (w = 0; w < width; w++) {
			*ptr = palette[*ptr][0];
			ptr++;
			*ptr = palette[*ptr][1];
			ptr++;
			*ptr = palette[*ptr][2];
			ptr++;
			*ptr = alpha_table[*ptr];
			ptr++;
		}
	}
}
#endif

static int load_png_file(const char *filename, png_byte **bitmap, png_uint_32 *width, png_uint_32 *height)
{
	int result = -1;
	FILE *fp = NULL;
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	png_byte **row_pointers = NULL;
	int h;
	unsigned char header[8];
	int bit_depth;
	int color_type;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: %d: ファイルオープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (fread(header, sizeof(header), 1, fp) != 1) {
		printf("ERROR: %d: 読み込みエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (png_sig_cmp(header, 0, sizeof(header)) != 0) {
		printf("ERROR: PNGファイルではありません: %s\n", filename);
		goto ERROR_EXIT;
	}

	if ((png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL) {
		printf("ERROR: png_create_read_struct()エラー\n");
		goto ERROR_EXIT;
	}

	if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
		printf("ERROR: png_create_info_struct()エラー\n");
		goto ERROR_EXIT;
	}

	if (setjmp(png_jmpbuf(png_ptr))) {
		printf("ERROR: PNGファイル読み込みエラー\n");
		goto ERROR_EXIT;
	}

	png_init_io(png_ptr, fp);

	png_set_sig_bytes(png_ptr, sizeof(header));

	png_read_info(png_ptr, info_ptr);

	png_get_IHDR(png_ptr,
				 info_ptr,
				 width,
				 height,
				 &bit_depth,
				 &color_type,
				 NULL,
				 NULL,
				 NULL);

	printf("%s\n", filename);
	printf("\twidth       = %d\n", *width);
	printf("\theight      = %d\n", *height);

	if (bit_depth != 8 || (color_type != PNG_COLOR_TYPE_RGB_ALPHA && color_type != PNG_COLOR_TYPE_RGB)) {
		printf("PNG画像の色数が違います\n");
		goto ERROR_EXIT;
	} else if (color_type == PNG_COLOR_TYPE_RGB) {
		printf("WARNING: アルファなし画像です: %s\n", filename);
	}

	if ((*bitmap = (png_byte *)malloc((*height) * (*width) * 4)) == NULL) {
		printf("ERROR: memory exhasted!\n");
		goto ERROR_EXIT;
	}

	if ((row_pointers = (png_byte **)malloc(sizeof(png_byte *) * (*height))) == NULL) {
		printf("ERROR: memory exhasted!\n");
		goto ERROR_EXIT;
	}

	for (h = 0; h < *height; h++) {
		row_pointers[h] = (*bitmap) + h * (*width) * 4;
	}

	png_read_image(png_ptr, row_pointers);

	png_read_end(png_ptr, NULL);

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (png_ptr != NULL) {
		png_destroy_read_struct(&png_ptr, (info_ptr != NULL ? &info_ptr : (png_infopp)NULL), (png_infopp)NULL);
	}

	if (fp != NULL) {
		fclose(fp);
	}

	if (row_pointers != NULL) {
		free(row_pointers);
	}

	if (result == 0 && color_type == PNG_COLOR_TYPE_RGB) {
		for (h = 0; h < *height; h++) {
			png_byte *src = (*bitmap) + h * (*width) * 4 + (*width) * 3 - 1;
			png_byte *des = (*bitmap) + h * (*width) * 4 + (*width) * 4 - 1;
			int w = *width;
			while (0 < w--) {
				*des-- = 0xFF;		/* alpha */
				*des-- = *src--;	/* blue  */
				*des-- = *src--;	/* green */
				*des-- = *src--;	/* red   */
			}
		}
	}

	if (result < 0) {
		if (*bitmap != NULL) {
			free(*bitmap);
		}
		*bitmap = NULL;
	}

	return result;
}

#ifndef PICKUP_COLOR
static int save_png_file(const char *filename, const png_byte *bitmap, png_uint_32 width, png_uint_32 height)
{
	int result = -1;
	FILE *fp = NULL;
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	const png_byte *row;
	int h;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: %d: ファイルオープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if ((png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL) {
		printf("ERROR: png_create_write_struct()エラー\n");
		goto ERROR_EXIT;
	}

	if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
		printf("ERROR: png_create_info_struct()エラー\n");
		goto ERROR_EXIT;
	}

	if (setjmp(png_jmpbuf(png_ptr))) {
		printf("ERROR: PNGファイル作成エラー\n");
		goto ERROR_EXIT;
	}

	png_set_compression_level(png_ptr, 9);

	png_init_io(png_ptr, fp);

	png_set_IHDR(png_ptr,
				 info_ptr,
				 width,
				 height,
				 8,
				 PNG_COLOR_TYPE_RGB_ALPHA,
				 PNG_INTERLACE_NONE,
				 PNG_COMPRESSION_TYPE_DEFAULT,
				 PNG_FILTER_TYPE_DEFAULT);

	png_write_info(png_ptr, info_ptr);

	row = bitmap;
	for (h = 0; h < height; h++) {
		png_write_row(png_ptr, row);
		row += width * 4;
	}

	png_write_end(png_ptr, NULL);

	fclose(fp);
	fp = NULL;

	result = 0;

ERROR_EXIT:
	if (png_ptr != NULL) {
		png_destroy_write_struct(&png_ptr, (info_ptr != NULL ? &info_ptr : (png_infopp)NULL));
	}

	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}
#endif

/* EOF */
