#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "file_utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


int read_next_line(char **buffer, size_t *size, FILE *fp)
{
	int result = -1;
	int result2 = 0;
	char *ptr = *buffer;
	const int expand_size = 10240;

	if (ptr == NULL) {
		*size = expand_size;
		if ((*buffer = ptr = (char *)malloc(*size)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}
	}

	if (fgets(ptr, *size, fp) != NULL) {
		size_t offset = strlen(ptr);

		while (0 < offset && ptr[offset - 1] != '\r' && ptr[offset - 1] != '\n') {
			*size += expand_size;
			if ((*buffer = ptr = (char *)realloc(ptr, *size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			if (fgets(ptr + offset, *size - offset, fp) == NULL) {
				break;
			}

			offset += strlen(ptr + offset);
		}

		result2 = 1;
	}

	if (ferror(fp)) {
		goto ERROR_EXIT;
	}

	result = result2;

ERROR_EXIT:
	if (*buffer == NULL) {
		*size = 0;
	}

	return result;
}

int read_append_next_line(char **buffer, size_t *size, FILE *fp)
{
	int result = -1;
	char *work_buf = NULL;
	size_t work_buf_size = 0;
	size_t alloc_size;

	if (*buffer == NULL || **buffer == 0) {
		return read_next_line(buffer, size, fp);
	}

	if ((result = read_next_line(&work_buf, &work_buf_size, fp)) <= 0) {
		/* error or eof */
		return result;
	}

	alloc_size = strlen(*buffer) + strlen(work_buf) + 1;
	if (*size < alloc_size) {
		if ((*buffer = (char *)realloc(*buffer, alloc_size)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			result = -1;
			goto ERROR_EXIT;
		}
		*size = alloc_size;
	}

	strcat(*buffer, work_buf);

ERROR_EXIT:
	if (work_buf != NULL) {
		free(work_buf);
	}

	return result;
}

int make_directory(const char *path, int path_is_file)
{
	int result = -1;
	char *work_buffer = NULL;
	char *dir = NULL;
	char *name = NULL;
	int is_fullpath = 0;
	struct stat st;
#ifdef WIN32
	const char *STR_PATH_SEP = "\\";
	const char CHAR_PATH_SEP = '\\';
#else
	const char *STR_PATH_SEP = "/";
	const char CHAR_PATH_SEP = '/';
#endif


	if (path_is_file == 0) {
		if (stat(path, &st) == 0 && S_ISDIR(st.st_mode)) {
			goto SUCCESS_EXIT;
		}
	}

	if ((work_buffer = strdup(path)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

#ifdef WIN32
	convert_slash_to_backslash(work_buffer);
#else
	convert_backslash_to_slash(work_buffer);
#endif

	if (path_is_file) {
		char *ptr;

		if ((ptr = strrchr(work_buffer, CHAR_PATH_SEP)) == NULL || ptr == work_buffer) {
			goto SUCCESS_EXIT;
		}

		*ptr = 0;
#ifdef WIN32
		if (strlen(work_buffer) == 2 && work_buffer[1] == ':') {
			goto SUCCESS_EXIT;
		}
#endif

		if (stat(work_buffer, &st) == 0 && S_ISDIR(st.st_mode)) {
			goto SUCCESS_EXIT;
		}
	}

	if ((dir = (char *)malloc(strlen(path) + 1)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}
	dir[0] = 0;

	if (work_buffer[0] == CHAR_PATH_SEP) {
		is_fullpath = 1;
	}

	name = strtok(work_buffer, STR_PATH_SEP);
	if (name == NULL) {
		printf("ERROR: illegal path: %s\n", path);
		goto ERROR_EXIT;
	}

	while (name != NULL) {
		if (dir[0] == 0 && is_fullpath == 0) {
			strcpy(dir, name);
		} else {
			strcat(strcat(dir, STR_PATH_SEP), name);
		}

#ifdef WIN32
		if (strlen(dir) == 2 && dir[1] == ':') {
			continue;
		}
#endif
		if (stat(dir, &st) != 0) {
			int res;
#ifdef WIN32
			res = mkdir(dir);
#else
			res = mkdir(dir, 0700);
#endif
			if (res < 0) {
				printf("ERROR: %d: can't make directory: %s\n", errno, dir);
				goto ERROR_EXIT;
			}
		}

		name = strtok(NULL, STR_PATH_SEP);
	}

SUCCESS_EXIT:
	result = 0;

ERROR_EXIT:
	if (work_buffer != NULL) {
		free(work_buffer);
	}
	return result;
}

void convert_slash_to_backslash(char *str)
{
	if (str == NULL) {
		return;
	}

	while (*str != 0) {
		if (*str == '/') {
			*str = '\\';
		}
		str++;
	}
}

void convert_backslash_to_slash(char *str)
{
	if (str == NULL) {
		return;
	}

	while (*str != 0) {
		if (*str == '\\') {
			*str = '/';
		}
		str++;
	}
}

/* EOF */
