#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>


#include "utils.h"
#include "file_utils.h"
#include "font_info_sub.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static void usage(void);
static void free_rect_file_table(struct rect_file_table_t *rect_file_table);
static int write_rect_file_list(const struct rect_file_table_t *rect_file_table, const char *filename);
static int write_rect_file_main(const struct rect_file_table_t *rect_file_table);
static int write_rect_file(const struct font_charset_t *charset, const char *filename);


int main(int argc, char **argv)
{
	int num_cmd_params = 3;
	const char *font_file = NULL;
	const char *list_file = NULL;
	const char *out_dir = NULL;

	struct font_table_t font_table;
	struct rect_file_table_t rect_file_table;

	memset(&font_table, 0, sizeof(font_table));
	memset(&rect_file_table, 0, sizeof(rect_file_table));

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 3) {
				font_file = *argv;
			} else if (num_cmd_params == 2) {
				list_file = *argv;
			} else {
				out_dir = *argv;
			}
			num_cmd_params--;
		}
	}

	if (font_file == NULL || list_file == NULL || out_dir == NULL) {
		usage();
		return 1;
	}

	if (read_font_info(&font_table, font_file) < 0) {
		printf("ERROR: フォント情報を読み込めません\n");
		return 1;
	}

	if (convert_font_rect(&rect_file_table, &font_table) < 0) {
		printf("ERROR: フォントから座標を取得できません\n");
		return 1;
	}

	{
		struct stat st;
		int i;

		if (stat(out_dir, &st) == 0) {
			if (!S_ISDIR(st.st_mode)) {
				printf("ERROR: ディレクトリではありません: %s\n", out_dir);
				return 1;
			}
		} else {
#ifdef WIN32
			if (mkdir(out_dir) < 0)
#else
			if (mkdir(out_dir, 0700) < 0)
#endif
			{
				printf("ERROR: %d: ディレクトリを作成できません: %s\n", errno, out_dir);
				return 1;
			}
		}

		for (i = 0; i < rect_file_table.size; i++) {
			char name[32];
			int len;
			char *rect_file;
#ifdef WIN32
			const char *separator = "\\";
#else
			const char *separator = "/";
#endif

			snprintf(name, sizeof(name), "%05d.txt", rect_file_table.table[i].index);
			len = strlen(out_dir) + 1 + strlen(name) + 1;
			if ((rect_file = (char *)malloc(len)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				return 1;
			}

			strcat(strcat(strcpy(rect_file, out_dir), separator), name);
			rect_file_table.table[i].rect_file = rect_file;
		}
	}

	if (write_rect_file_list(&rect_file_table, list_file) < 0) {
		printf("ERROR: 座標ファイル一覧を作成できません\n");
		return 1;
	}

	if (write_rect_file_main(&rect_file_table) < 0) {
		printf("ERROR: 座標ファイルを作成できません\n");
		return 1;
	}

	free_rect_file_table(&rect_file_table);
	free_font_info(&font_table);

	return 0;
}

static void usage(void)
{
	printf("使用方法: font_char_rect [オプション] フォント情報ファイル 座標ファイル一覧\n");
	printf("                                                               出力ディレクトリ\n");
	printf("機    能: フォント情報ファイルを読み込み座標ファイルを指定ディレクトリに作成し\n");
	printf("          ます\n");
	printf("オプション:\n");
	printf("    -?|--help               使用方法を表示して終了します\n");
}

static void free_rect_file_table(struct rect_file_table_t *rect_file_table)
{
	if (rect_file_table->table != NULL) {
		int i;

		for (i = 0; i < rect_file_table->size; i++) {
			if (rect_file_table->table[i].rect_file != NULL) {
				free(rect_file_table->table[i].rect_file);
			}
			free_font_charset(&(rect_file_table->table[i].charset));
		}

		free(rect_file_table->table);
	}

	memset(rect_file_table, 0, sizeof(*rect_file_table));
}

static int write_rect_file_list(const struct rect_file_table_t *rect_file_table, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	int i;

	if ((fp = fopen(filename, "wt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	for (i = 0; i < rect_file_table->size; i++) {
		if (fprintf(fp, "%05d\t%s\n", rect_file_table->table[i].index, rect_file_table->table[i].rect_file) < 0) {
			printf("ERROR: 書き込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int write_rect_file_main(const struct rect_file_table_t *rect_file_table)
{
	int i;

	for (i = 0; i < rect_file_table->size; i++) {
		if (write_rect_file(&(rect_file_table->table[i].charset), rect_file_table->table[i].rect_file) < 0) {
			printf("ERROR: 文字座標ファイルを作成できません: %s\n", rect_file_table->table[i].rect_file);
			return -1;
		}
	}

	return 0;
}

static int write_rect_file(const struct font_charset_t *charset, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	int i;
	const char *header = "# 文字\t左端X座標\t上端Y座標\t右端X座標\t下端Y座標";

	if ((fp = fopen(filename, "wt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (fprintf(fp, "%s\n", header) < 0) {
		printf("ERROR: 書き込みエラー: %s\n", filename);
		goto ERROR_EXIT;
	}

	for (i = 0; i < charset->used; i++) {
		const struct font_char_rect_t *rect = &(charset->rect[i]);
		char buffer[32];
		int size = sizeof(buffer);

		code_to_string(buffer, &size, rect->code);
		if (fprintf(fp, "%s\t%d\t%d\t%d\t%d\n", buffer, rect->left, rect->top, rect->right, rect->bottom) < 0) {
			printf("ERROR: 書き込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

/* EOF */
