#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


#include "utils.h"
#include "file_utils.h"
#include "font_info_sub.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static int parse_font_no(char *src, int *font_no);
static int check_font_spacing(const struct font_charset_t *charset, const char *spacing);
static int parse_char_rect(char *src, unsigned long *code, int *len, int *left, int *top, int *right, int *bottom);
static int expand_font_charset(struct font_charset_t *charset);
static struct font_char_rect_t *find_char_rect(struct font_charset_t *charset, unsigned long code, int *index);
static int add_char_rect(struct font_charset_t *charset, unsigned long code, int len, int left, int top, int right, int bottom);


void free_font_info(struct font_table_t *font_table)
{
	int i;

	if (font_table->t141 == NULL) {
		return;
	}

	for (i = 0; i < font_table->size; i++) {
		if (font_table->t141[i].t12 != NULL) {
			free(font_table->t141[i].t12);
		}
		if (font_table->t141[i].t248.t7 != NULL) {
			free(font_table->t141[i].t248.t7);
		}
		if (font_table->t141[i].t143 != NULL) {
			if (--(font_table->t141[i].t143->link_count) <= 0) {
				if (font_table->t141[i].t143->t22 != NULL) {
					free(font_table->t141[i].t143->t22);
				}
				free(font_table->t141[i].t143);
			}
		}
		if (font_table->t141[i].t249 != NULL) {
			free(font_table->t141[i].t249);
		}
		if (font_table->t141[i].t703 != NULL) {
			free(font_table->t141[i].t703);
		}
	}

	free(font_table->t141);

	memset(font_table, 0, sizeof(*font_table));
}

const struct font_t141_t *find_font(const struct font_table_t *font_table, int t13)
{
	int i;

	for (i = 0; i < font_table->size; i++) {
		if (font_table->t141[i].t13 == t13) {
			return &(font_table->t141[i]);
		}
	}

	return NULL;
}

int read_font_info(struct font_table_t *font_table, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	char *src;
	int font_no = -1;
	int line = 0;
	int found_charset = 0;
	int found_char = 0;
	int found_spacing = 0;
	struct font_t141_t *cur_font = NULL;
	struct font_charset_t charset;

	enum font_token_id {
		TOKEN_FONT_NAME,
		TOKEN_FONT_PATH,
		TOKEN_FONT_SPACING,
		TOKEN_FONT_CHARSET,
		TOKEN_FONT_CHAR
	} token_id;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	memset(&charset, 0, sizeof(charset));

	if ((fp = fopen(filename, "rt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, fp)) < 0) {
			printf("ERROR: 読み込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;
		src = buffer;

		if (line == 1) {
			src = skip_bom(src);
		}

		src = skip_space(src);

		if (*src == '#' || *src == ';') {
			continue;
		}

		if (*src == '\r' || *src == '\n') {
			continue;
		}

		if (*src == '[') {
			size_t alloc_size;

			if (parse_font_no(src, &font_no) < 0) {
				printf("ERROR: 不正なフォント番号です: %s: %d行目\n", filename, line);
				goto ERROR_EXIT;
			}

			if (cur_font != NULL) {
				if (found_spacing && check_font_spacing(&charset, cur_font->t703) < 0) {
					printf("ERROR: 不正なspacing定義です: %s: %d行目\n", filename, found_spacing);
					goto ERROR_EXIT;
				}
				if (charset.size > 0) {
					if (convert_font_charset(cur_font, &charset) < 0) {
						goto ERROR_EXIT;
					}
					free_font_charset(&charset);
				}

				if (cur_font->t143 == NULL && cur_font->ref_t13 < 0) {
					printf("ERROR: フォント%dにcharとcharsetのどちらも定義されていません: %s: %d行目\n", cur_font->t13, filename, line);
					goto ERROR_EXIT;
				}
				if (cur_font->t12 == NULL) {
					printf("WARNING: フォント%dにnameが定義されていません: %s: %d行目\n", cur_font->t13, filename, line);
				}
				if (cur_font->t248.t7 == NULL) {
					printf("ERROR: フォント%dにpathが定義されていません: %s: %d行目\n", cur_font->t13, filename, line);
					goto ERROR_EXIT;
				}
				if (cur_font->t703 == NULL && cur_font->ref_t13 < 0) {
					printf("WARNING: フォント%dにspacingが定義されていません: %s\n", cur_font->t13, filename);
				}
			}

			if (find_font(font_table, font_no) != NULL) {
				printf("ERROR: フォント%dは定義済みです: %s: %d行目\n", font_no, filename, line);
				goto ERROR_EXIT;
			}

			alloc_size = sizeof(struct font_t141_t) * (font_table->size + 1);
			if (font_table->size == 0) {
				font_table->t141 = (struct font_t141_t *)malloc(alloc_size);
			} else {
				font_table->t141 = (struct font_t141_t *)realloc(font_table->t141, alloc_size);
			}

			if (font_table->t141 == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			memset(&(font_table->t141[font_table->size]), 0, sizeof(struct font_t141_t));
			font_table->size++;

			cur_font = &(font_table->t141[font_table->size - 1]);
			cur_font->t13 = font_no;
			cur_font->ref_t13 = -1;

			found_charset = 0;
			found_char = 0;
			found_spacing = 0;

			continue;
		} else {
			char *value;

			if (font_no < 0) {
				printf("ERROR: 構文エラー: %s: %d行目\n", filename, line);
				goto ERROR_EXIT;
			}

			if (strncasecmp(src, "name", 4) == 0) {
				token_id = TOKEN_FONT_NAME;
				src += 4;
			} else if (strncasecmp(src, "path", 4) == 0) {
				token_id = TOKEN_FONT_PATH;
				src += 4;
			} else if (strncasecmp(src, "spacing", 7) == 0) {
				token_id = TOKEN_FONT_SPACING;
				src += 7;
			} else if (strncasecmp(src, "charset", 7) == 0) {
				token_id = TOKEN_FONT_CHARSET;
				src += 7;
			} else if (strncasecmp(src, "char", 4) == 0) {
				token_id = TOKEN_FONT_CHAR;
				src += 4;
			} else {
				printf("ERROR: 構文エラー: %s: %d行目\n", filename, line);
				goto ERROR_EXIT;
			}

			src = skip_space(src);
			if (*src != '=') {
				printf("ERROR: 構文エラー: %s: %d行目\n", filename, line);
				goto ERROR_EXIT;
			}

			src++;
			src = skip_space(src);

			value = src;
			trim_right(src);

			if (token_id == TOKEN_FONT_NAME) {
				if (cur_font->t12 != NULL) {
					printf("ERROR: nameは定義済です: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if ((cur_font->t12 = strdup(value)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}
			} else if (token_id == TOKEN_FONT_PATH) {
				if (cur_font->t248.t7 != NULL) {
					printf("ERROR: pathは定義済です: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if ((cur_font->t248.t7 = strdup(value)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}
			} else if (token_id == TOKEN_FONT_SPACING) {
				if (found_charset) {
					printf("ERROR: charsetとspacingは同じフォントに混在できません: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if (cur_font->t703 != NULL) {
					printf("ERROR: spacingは定義済です: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if ((cur_font->t703 = escape_text(value)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}

				found_spacing = line;
			} else if (token_id == TOKEN_FONT_CHARSET) {
				int ref_no = -1;

				if (found_charset) {
					printf("ERROR: charsetは同じフォントに複数回使用できません: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if (found_char) {
					printf("ERROR: charsetとcharは同じフォントに混在できません: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if (found_spacing) {
					printf("ERROR: charsetとspacingは同じフォントに混在できません: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}

				if (parse_font_no(value, &ref_no) < 0) {
					printf("ERROR: 不正なフォント番号です: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}
				if (cur_font->t13 == ref_no) {
					printf("ERROR: 参照元と参照先が同じフォント番号です: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}

				cur_font->ref_t13 = ref_no;

				found_charset = line;
			} else {
				int len;
				unsigned long code;
				int left;
				int top;
				int right;
				int bottom;

				if (found_charset) {
					printf("ERROR: charsetとcharは同じフォントに混在できません: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}

				if (parse_char_rect(value, &code, &len, &left, &top, &right, &bottom) < 0) {
					printf("ERROR: 不正な文字定義です: %s: %d行目\n", filename, line);
					goto ERROR_EXIT;
				}

				if (find_char_rect(&charset, code, NULL) != NULL) {
					char buffer[32];
					int size = sizeof(buffer);
					code_to_string(buffer, &size, code);
					printf("WARNING: 登録済の文字です: %s: %d行目: char=%s\n", filename, line, buffer);
				}

				if (add_char_rect(&charset, code, len, left, top, right, bottom) < 0) {
					goto ERROR_EXIT;
				}

				found_char = line;
			}
		}
	}

	if (cur_font != NULL) {
		if (found_spacing && check_font_spacing(&charset, cur_font->t703) < 0) {
			printf("ERROR: 不正なspacing定義です: %s: %d行目\n", filename, found_spacing);
			goto ERROR_EXIT;
		}
		if (charset.size > 0) {
			if (convert_font_charset(cur_font, &charset) < 0) {
				goto ERROR_EXIT;
			}
			free_font_charset(&charset);
		}

		if (cur_font->t143 == NULL && cur_font->ref_t13 < 0) {
			printf("ERROR: フォント%dにcharとcharsetのどちらも定義されていません: %s\n", cur_font->t13, filename);
			goto ERROR_EXIT;
		}
		if (cur_font->t12 == NULL) {
			printf("WARNING: フォント%dにnameが定義されていません: %s\n", cur_font->t13, filename);
		}
		if (cur_font->t248.t7 == NULL) {
			printf("ERROR: フォント%dにpathが定義されていません: %s\n", cur_font->t13, filename);
			goto ERROR_EXIT;
		}
		if (cur_font->t703 == NULL && cur_font->ref_t13 < 0) {
			printf("WARNING: フォント%dにspacingが定義されていません: %s\n", cur_font->t13, filename);
		}
	}

	if (font_table->size == 0) {
		printf("ERROR: フォント定義がありません: %s\n", filename);
		goto ERROR_EXIT;
	}

	/* 参照先の存在確認 */
	{
		int i;

		for (i = 0; i < font_table->size; i++) {
			struct font_t141_t *cur_font = &(font_table->t141[i]);
			const struct font_t141_t *ref_font = NULL;

			/* 参照先フォントID */
			if (cur_font->ref_t13 < 0) {
				continue;
			}

			if ((ref_font = find_font(font_table, cur_font->ref_t13)) == NULL) {
				printf("ERROR: フォント%dの参照先フォント%dは定義されていません: %s\n", cur_font->t13, cur_font->ref_t13, filename);
				goto ERROR_EXIT;
			}

			/* 文字一覧 */
			if (ref_font->t249 == NULL) {
				cur_font->t249 = NULL;
			} else if ((cur_font->t249 = strdup(ref_font->t249)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			/* spacing */
			if (ref_font->t703 == NULL) {
				cur_font->t703 = NULL;
			} else if ((cur_font->t703 = strdup(ref_font->t703)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			/* 文字矩形領域座標の配列 */
			cur_font->t143 = ref_font->t143;
			cur_font->t143->link_count++;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static int parse_font_no(char *src, int *font_no)
{
	char *endptr = NULL;
	*font_no = -1;

	src = skip_space(src);
	if (*src != '[') {
		return -1;
	}
	src++;

	src = skip_space(src);
	*font_no = strtoul(src, &endptr, 10);

	if (endptr == NULL || src == endptr) {
		return -1;
	}

	src = skip_space(endptr);
	if (*endptr != ']') {
		return -1;
	}
	src++;

	src = skip_space(src);
	src = skip_crlf(src);
	if (*src != 0) {
		return -1;
	}

	return 0;
}

static int check_font_spacing(const struct font_charset_t *charset, const char *spacing)
{
	int count;

	if (spacing == NULL || *spacing == 0) {
		return 0;
	}

	for (count = 0; *spacing != 0; count++) {
		int len;
		unsigned long code1;
		unsigned long code2;
		int i;
		int found;

		if (0 < count) {
			if (*spacing != '|') {
				return -1;
			}
			spacing++;
		}

		if ((len = get_next_utf8_char(&code1, spacing, 1)) < 0) {
			return -1;
		}
		spacing += len;

		if (*spacing == 0 || (len = get_next_utf8_char(&code2, spacing, 1)) < 0) {
			return -1;
		}
		spacing += len;

		if (*spacing != '=') {
			return -1;
		}
		spacing++;

		if (*spacing == '-') {
			spacing++;
		}
		if (*spacing < '0' || '9' < *spacing) {
			return -1;
		}
		spacing++;
		while ('0' <= *spacing && *spacing <= '9') {
			spacing++;
		}

		for (found = 0, i = 0; i < charset->used && found != 0x03; i++) {
			if (charset->rect[i].code == code1) {
				found |= 0x01;
			}
			if (charset->rect[i].code == code2) {
				found |= 0x02;
			}
		}
		if ((found & 0x01) == 0) {
			char buffer[32];
			int size = sizeof(buffer);
			code_to_string(buffer, &size, code1);
			printf("WARNING: 未定義の文字%sがspacingで指定されています\n", buffer);
		}
		if (code1 != code2 && (found & 0x02) == 0) {
			char buffer[32];
			int size = sizeof(buffer);
			code_to_string(buffer, &size, code2);
			printf("WARNING: 未定義の文字%sがspacingで指定されています\n", buffer);
		}
	}

	return 0;
}

static int parse_char_rect(char *src, unsigned long *code, int *len, int *left, int *top, int *right, int *bottom)
{
	const char *separater = " \t\r\n";
	char *token;
	int i;
	int value[4];
	char *endptr;

	if ((token = strtok(src, separater)) == NULL) {
		return -1;
	}
	if ((*len = get_next_utf8_char(code, token, 0)) < 0 || strlen(token) != *len) {
		printf("ERROR: 不正な文字指定です\n");
		return -1;
	}

	for (i = 0; i < 4; i++) {
		if ((token = strtok(NULL, separater)) == NULL) {
			return -1;
		}

		value[i] = strtoul(token, &endptr, 10);
		if (endptr == NULL || *endptr != 0 || value[i] < 0) {
			return -1;
		}
	}

	*left   = value[0];
	*top    = value[1];
	*right  = value[2];
	*bottom = value[3];

	if (*right < *left || *bottom < *top) {
		printf("ERROR: 不正な矩形範囲です\n");
		return -1;
	}

	if (strtok(NULL, separater) != NULL) {
		return -1;
	}

	return 0;
}

int add_rect_file(struct rect_file_table_t *rect_file_table, int index, char *rect_file)
{
	const struct rect_file_info_t *rect_file_info;
	size_t alloc_size;

	if ((rect_file_info = find_rect_file_info(rect_file_table, index)) != NULL) {
		return 0;
	}

	alloc_size = sizeof(struct rect_file_info_t) * (rect_file_table->size + 1);
	if (rect_file_table->size == 0) {
		rect_file_table->table = (struct rect_file_info_t *)malloc(alloc_size);
	} else {
		rect_file_table->table = (struct rect_file_info_t *)realloc(rect_file_table->table, alloc_size);
	}
	if (rect_file_table->table == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}

	memset(&(rect_file_table->table[rect_file_table->size]), 0, sizeof(struct rect_file_info_t));

	rect_file_table->table[rect_file_table->size].index        = index;
	rect_file_table->table[rect_file_table->size].rect_file    = rect_file;

	rect_file_table->table[rect_file_table->size].charset.size = 0;
	rect_file_table->table[rect_file_table->size].charset.used = 0;
	rect_file_table->table[rect_file_table->size].charset.charset_str_len = 0;
	rect_file_table->table[rect_file_table->size].charset.rect = NULL;

	rect_file_table->size++;

	return 1;
}

struct rect_file_info_t *find_rect_file_info(struct rect_file_table_t *rect_file_table, int index)
{
	int i;
	struct rect_file_info_t *info = rect_file_table->table;

	for (i = 0; i < rect_file_table->size; i++, info++) {
		if (info->index == index) {
			return info;
		}
	}

	return NULL;
}

void free_font_charset(struct font_charset_t *charset)
{
	if (charset->rect != NULL) {
		free(charset->rect);
	}

	memset(charset, 0, sizeof(*charset));
}

static int expand_font_charset(struct font_charset_t *charset)
{
	const int expand_size = 200;
	size_t alloc_size = sizeof(struct font_char_rect_t) * (charset->size + expand_size);

	if (charset->size == 0) {
		charset->rect = (struct font_char_rect_t *)malloc(alloc_size);
	} else {
		charset->rect = (struct font_char_rect_t *)realloc(charset->rect, alloc_size);
	}

	if (charset->rect == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}

	memset(&(charset->rect[charset->size]), 0, sizeof(struct font_char_rect_t) * expand_size);
	charset->size += expand_size;

	return 0;
}

static struct font_char_rect_t *find_char_rect(struct font_charset_t *charset, unsigned long code, int *index)
{
	int i = 0;

	if (index != NULL && 0 <= *index) {
		i = *index;
		*index = -1;
	}

	for ( ; i < charset->used; i++) {
		if (charset->rect[i].code == code) {
			if (index != NULL) {
				*index = i;
			}
			return &(charset->rect[i]);
		}
	}

	return NULL;
}

static int add_char_rect(struct font_charset_t *charset, unsigned long code, int len, int left, int top, int right, int bottom)
{
	if (charset->size <= charset->used) {
		if (expand_font_charset(charset) < 0) {
			return -1;
		}
	}

	charset->charset_str_len += len;
	charset->rect[charset->used].code   = code;
	charset->rect[charset->used].left   = left;
	charset->rect[charset->used].top    = top;
	charset->rect[charset->used].right  = right;
	charset->rect[charset->used].bottom = bottom;
	charset->used++;

	return 0;
}

int read_char_rect(struct font_charset_t *charset, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	char *src;
	int line = 0;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	if ((fp = fopen(filename, "rt")) == NULL) {
		if (errno == ENOENT) {
			result = 0;
			printf("INFORMATION: %s は存在しません ... スキップ\n", filename);
		} else {
			printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		}
		goto ERROR_EXIT;
	}

	while (1) {
		int len;
		unsigned long code;
		int left;
		int top;
		int right;
		int bottom;

		if ((read_status = read_next_line(&buffer, &buffer_size, fp)) < 0) {
			printf("ERROR: 読み込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;
		src = buffer;

		if (line == 1) {
			/*src = skip_bom(src);*/
			/* 1行目は読み飛ばす */
			continue;
		}

		if (parse_char_rect(src, &code, &len, &left, &top, &right, &bottom) < 0) {
			printf("ERROR: 不正な文字定義です: %s: %d行目\n", filename, line);
			goto ERROR_EXIT;
		}

		if (code == 0) {
			continue;
		}

		if (find_char_rect(charset, code, NULL) != NULL) {
			char buffer[32];
			int size = sizeof(buffer);
			code_to_string(buffer, &size, code);
			printf("WARNING: 登録済の文字です: %s: %d行目: char=%s\n", filename, line, buffer);
		}

		if (add_char_rect(charset, code, len, left, top, right, bottom) < 0) {
			goto ERROR_EXIT;
		}
	}

	result = 1;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

int convert_font_charset(struct font_t141_t *t141, const struct font_charset_t *charset)
{
	size_t alloc_size;
	struct font_t143_t *t143;
	struct font_t22_t *t22;
	int i;
	char *str;
	int str_buffer_size;

	alloc_size = str_buffer_size = charset->charset_str_len + 1;
	if ((t141->t249 = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}
	memset(t141->t249, 0, alloc_size);

	alloc_size = sizeof(struct font_t143_t);
	if ((t143 = (struct font_t143_t *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}
	memset(t143, 0, alloc_size);
	t141->t143 = t143;

	t143->link_count = 1;
	t143->size = charset->used;

	alloc_size = sizeof(struct font_t22_t) * (t143->size);
	if ((t22 = (struct font_t22_t *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}
	memset(t22, 0, alloc_size);
	t143->t22 = t22;

	str = t141->t249;
	for (i = 0; i < charset->used; i++) {
		if ((str = code_to_string(str, &str_buffer_size, charset->rect[i].code)) == NULL) {
			return -1;
		}
		t22[i].t3 = charset->rect[i].top;
		t22[i].t4 = charset->rect[i].left;
		t22[i].t5 = charset->rect[i].bottom - charset->rect[i].top + 1;
		t22[i].t6 = charset->rect[i].right - charset->rect[i].left + 1;
	}

	if (check_escape_chars(t141->t249)) {
		char *str2;
		if ((str2 = escape_text(t141->t249)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			return -1;
		}

		free(t141->t249);
		t141->t249 = str2;
	}

	return 0;
}

int convert_font_rect(struct rect_file_table_t *rect_file_table, const struct font_table_t *font_table)
{
	int result = -1;
	int i;

	for (i = 0; i < font_table->size; i++) {
		const struct font_t141_t *t141 = &(font_table->t141[i]);
		const struct font_t143_t *t143 = t141->t143;
		const char *path = t141->t248.t7;
		const char *src = t141->t249;
		int index;
		int j;
		struct rect_file_info_t *info;
		struct font_charset_t *charset;

		if ((index = get_path_index(path)) < 0) {
			if (0 <= t141->ref_t13) {
#if 0
				printf("WARNING: path指定がありません: フォント%d, 参照先=%d\n", t141->t13, t141->ref_t13);
#endif
				continue;
			}
			printf("ERROR: 不正なpathです: フォント%d, path=%s\n", t141->t13, path);
			goto ERROR_EXIT;
		}

		if (add_rect_file(rect_file_table, index, NULL) < 0) {
			goto ERROR_EXIT;
		}

		if ((info = find_rect_file_info(rect_file_table, index)) == NULL) {
			printf("BUG: 座標ファイル情報が見つかりません: index=%05d\n", index);
			goto ERROR_EXIT;
		}
		charset = &(info->charset);

		for (j = 0; j < t143->size; j++) {
			const struct font_t22_t *t22 = &(t143->t22[j]);
			unsigned long code;
			int len;
			int rect_index = 0;
			int left   = t22->t4;
			int top    = t22->t3;
			int right  = t22->t4 + t22->t6 - 1;
			int bottom = t22->t3 + t22->t5 - 1;
			const struct font_char_rect_t *found_rect = NULL;

			if ((len = get_next_utf8_char(&code, src, 1)) < 0) {
				printf("ERROR: フォント%d: 文字セットに不正な文字が含まれています\n", t141->t13);
				goto ERROR_EXIT;
			}
			if (len == 0) {
				printf("ERROR: フォント%d: 文字セットが不足しています\n", t141->t13);
				goto ERROR_EXIT;
			}

			rect_index = 0;
			while (1) {
				const struct font_char_rect_t *rect = find_char_rect(charset, code, &rect_index);
				if (rect == NULL) {
					if (found_rect != NULL) {
						char buffer[32];
						int size = sizeof(buffer);
						code_to_string(buffer, &size, code);
						printf("WARNING: 座標の違う同じ文字が登録されています: char=%s, path=%s\n", buffer, path);
					}

					if (add_char_rect(charset, code, len, left, top, right, bottom) < 0) {
						goto ERROR_EXIT;
					}

					break;
				}

				if (rect->left   == left &&
					rect->top    == top &&
					rect->right  == right &&
					rect->bottom == bottom) {
					/* 座標が同じなら登録しない */
					break;
				}

				found_rect = rect;
				rect_index++;
			}

			src += len;
		}
	}

	result = 0;

ERROR_EXIT:

	return result;
}

int get_path_index(const char *path)
{
	int index = -1;
	int len;
	char *endptr;

	if (path == NULL || *(path) == '\0') {
		return -1;
	}

	len = strlen(path);
	if (len < 13 || strncmp(path + len - 13, "#v#-01#", 7) != 0) {
		return -1;
	}

	index = strtol(path + len - 6, &endptr, 10);
	if (endptr != path + len - 1 || *endptr != '#') {
		return -1;
	}

	return index;
}

/* EOF */
