#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "utils.h"
#include "file_utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static void usage(void);
static int get_script(const char *in_file, const char *script_name, const char *out_file, int opt_all, int opt_list);
static void convert_br(char *src);


int main(int argc, char **argv)
{
	int num_cmd_params = 3;
	const char *xml_file = NULL;
	const char *script_name = NULL;
	const char *script_file = NULL;
	int opt_list = 0;
	int opt_all = 0;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-l") == 0 || strcmp(*argv, "--list") == 0) {
				opt_list = 1;
			} else if (strcmp(*argv, "-a") == 0 || strcmp(*argv, "--all") == 0) {
				opt_all = 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 3) {
				xml_file = *argv;
			} else if (num_cmd_params == 2) {
				script_name = *argv;
			} else {
				script_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (xml_file == NULL ||
		(opt_list == 0 && opt_all == 0 && script_name == NULL) ||
		(script_name != NULL && script_file == NULL)) {
		usage();
		return 1;
	}

	if (opt_all != 0 && script_name != NULL) {
		printf("WARNING: -a|--allオプション使用時はスクリプト名|IDと出力ファイル名が\n");
		printf("         指定されていても無視します\n");
		script_name = NULL;
		script_file = NULL;
	}

	if (get_script(xml_file, script_name, script_file, opt_all, opt_list) < 0) {
		printf("ERROR: スクリプトを抽出できません\n");
		return 1;
	}

	return 0;
}

static void usage(void)
{
	printf("使用方法: get_script [オプション] XMLファイル [スクリプト名|ID 出力ファイル]\n");
	printf("機    能: XMLファイルからスクリプトを抽出して保存します\n");
	printf("オプション:\n");
	printf("    -l|--list           スクリプトIDとスクリプト名の一覧を表示します\n");
	printf("    -a|--all            全てのスクリプトをscript.番号.txtという名前で\n");
	printf("                        保存します\n");
	printf("    -?|--help           使用方法を表示して終了します\n");
	printf("\n");
	printf("スクリプト名が-で始まる場合はスクリプトIDで指定してください\n");
}

static int get_script(const char *in_file, const char *script_name, const char *out_file, int opt_all, int opt_list)
{
	int result = -1;
	FILE *in_fp = NULL;
	FILE *out_fp = NULL;
	int found = 0;
	long line = 0;
	long id;
	char *ptr;
	char *name;
	char *script;
	const char bom[3] = { 0xEF, 0xBB, 0xBF };
	long script_id = -1;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	char name_buffer[256];

	if (opt_all) {
		printf("全スクリプト保存\n");
	} else if (script_name != NULL) {
		const char *p;
		for (p = script_name; *p != 0 && isdigit(*p); p++) ;
		if (*p == 0) {
			script_id = strtoul(script_name, NULL, 10);
		}
	}

	if ((in_fp = fopen(in_file, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, in_file);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
			printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;

		if ((ptr = strstr(buffer, "<T650 ")) == NULL) {
			/* スクリプト以外 */
			continue;
		}

		if ((ptr = strstr(ptr + 5, " T12=\"")) == NULL) {
			printf("ERROR: スクリプト名が見つかりません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		name = ptr + 6;
		if ((ptr = strstr(name, "\"")) == NULL) {
			printf("ERROR: スクリプト名が\"で終っていません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		if (name < ptr) {
			if (unescape_text(name, ptr) < 0) {
				printf("WARNING: 不正な文字が使用されています: %s(%ld)\n", in_file, line);
				goto ERROR_EXIT;
			}
		}

		if ((ptr = strstr(ptr + 1, " T13=\"")) == NULL) {
			printf("ERROR: スクリプトIDが見つかりません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		id = strtoul(ptr + 6, &ptr, 10);
		if (ptr == NULL || *ptr != '\"') {
			printf("ERROR: スクリプトIDの書式が違います: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		if (opt_list) {
			printf("%ld\t%s\n", id, name);
		}

		if (opt_all) {
			snprintf(name_buffer, sizeof(name_buffer), "script.%ld.txt", id);
			out_file = name_buffer;
		} else {
			if (script_name == NULL || !((0 <= script_id && id == script_id) || strcmp(name, script_name) == 0)) {
				continue;
			}
		}

		if ((ptr = strstr(ptr + 1, " T652=\"")) == NULL) {
			printf("ERROR: スクリプト本体が見つかりません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		script = ptr + 7;
		if ((ptr = strstr(script, "\"")) == NULL) {
			printf("ERROR: スクリプトが\"で終っていません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		if (script == ptr) {
			printf("WARNING: スクリプトの中身がありません: %s(%ld)\n\tid=%ld, name=%s\n", in_file, line, id, name);
			continue;
		}

		if (unescape_text(script, ptr) < 0) {
			printf("ERROR: 不正な文字が使用されています: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		convert_br(script);

		{
			const char *p;
			int space_only = 1;
			for (p = script; *p != '\0'; p++) {
				if (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n') {
					space_only = 0;
					break;
				}
			}
			if (space_only != 0) {
				printf("WARNING: スクリプトの中身は空白、改行のみです: %s(%ld)\n\tid=%ld, name=%s\n", in_file, line, id, name);
				continue;
			}
		}

		if ((out_fp = fopen(out_file, "wb")) == NULL) {
			printf("ERROR: %d: オープンエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		if (fwrite(bom, 3, 1, out_fp) != 1) {
			printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		if (fprintf(out_fp, "ID   = %ld\n", id) < 0) {
			printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		if (fprintf(out_fp, "name = %s\n", name) < 0) {
			printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		if (fprintf(out_fp, "%s", script) < 0) {
			printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		fclose(out_fp);
		out_fp = NULL;

		found++;

		if (opt_all == 0 && opt_list == 0) {
			break;
		}
	}

	if (script_name != NULL && !found) {
		printf("ERROR: スクリプトが見付かりません: %s\n", script_name);
	}

	result = 0;

ERROR_EXIT:
	if (in_fp != NULL) {
		fclose(in_fp);
	}
	if (out_fp != NULL) {
		fclose(out_fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static void convert_br(char *src)
{
	char *des = src;

	while (*src != 0) {
		if (*src == '<' && *(src + 1) == 'b' && *(src + 2) == 'r' && *(src + 3) == '/' && *(src + 4) == '>') {
			*des++ = '\n';
			src += 5;
		} else {
			*des++ = *src++;
		}
	}

	*des = 0;
}

/* EOF */
