/*
 * PNG画像のマスク処理はVISExtから移植しました。
 * VISExt 1.2.5174 Copyright (C) 2010-2014, Deniz Oezmen (http://oezmen.eu/)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>

#ifdef WIN32
#include "md5.h"
#define MD5_CTX md5_state_t
#define MD5Init(_p) md5_init(_p)
#define MD5Update(_p,_d,_n) md5_append(_p,(const md5_byte_t *)_d,_n)
#define MD5Final(_d,_p) md5_finish(_p,_d)
#else
#include <md5.h>
#endif

#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


#ifdef WIN32
#define PATH_DELIM	"\\"
#else
#define PATH_DELIM	"/"
#endif


static const unsigned char png_sig[8] = {
	0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A
};


static void usage(void);
static int mask_png_main(const char *in_dir, const char *out_dir);
static int check_png_name(const char *filename);
static int read_png_file(const char *dir, const char *filename, unsigned char **data, long *size);
static int write_png_file(const char *dir, int no, const unsigned char *data, long size);
static void mask_width_and_height(unsigned char *data, long size, const unsigned char *key);
static void make_png_key(unsigned char *key, const unsigned char *digest);


int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *in_dir = NULL;
	const char *out_dir = NULL;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				in_dir = *argv;
			} else {
				out_dir = *argv;
			}
			num_cmd_params--;
		}
	}

	if (in_dir == NULL || out_dir == NULL) {
		usage();
		return 1;
	}

	if (mask_png_main(in_dir, out_dir) < 0) {
		printf("ERROR: 変換エラーが発生しました\n");
		return 1;
	}

	return 0;
}

static void usage(void)
{
	printf("使用方法: mask_png [オプション] 入力DIR 出力DIR\n");
	printf("機    能: 画像ファイルをマスクします\n");
	printf("オプション:\n");
	printf("    -?|--help           使用方法を表示して終了します\n");
}

static int mask_png_main(const char *in_dir, const char *out_dir)
{
	int result = -1;
	struct stat st;
	DIR *dirp = NULL;
	struct dirent *entry;
	int no;
	const char *basename;
	int basename_len;
	MD5_CTX md5;
	unsigned char digest[16];
	unsigned char png_key[8];
	unsigned char *data = NULL;
	long size;

	if (stat(in_dir, &st) == 0) {
		if (!S_ISDIR(st.st_mode)) {
			printf("ERROR: ディレクトリではありません: %s\n", out_dir);
			goto ERROR_EXIT;
		}
	}

	if ((dirp = opendir(in_dir)) == NULL) {
		printf("ERROR: %d: ディレクトリオープンエラー: %s\n", errno, in_dir);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((entry = readdir(dirp)) == NULL) {
			if (errno != 0 && errno != ENOENT) {
				printf("ERROR: %d: ディレクトリを読めません: %s\n", errno, in_dir);
				goto ERROR_EXIT;
			}
			break;
		}
		if (check_png_name(entry->d_name) < 0) {
			continue;
		}

		if (read_png_file(in_dir, entry->d_name, &data, &size) < 0) {
			printf("ERROR: PNGファイルを読み込めません: %s%s%s\n", in_dir, PATH_DELIM, entry->d_name);
			goto ERROR_EXIT;
		}

		no = strtoul(entry->d_name, NULL, 10);
		basename = entry->d_name + 6;
		basename_len = strlen(basename) - 4;

		MD5Init(&md5);
		MD5Update(&md5, basename, basename_len);
		MD5Final(digest, &md5);

		printf("md5 = ");
		{ int i; for (i = 0; i < 16; i++) { printf("%02X", digest[i]); } }
		printf("\n");

		make_png_key(png_key, digest);

		mask_width_and_height(data, size, png_key);

		if (write_png_file(out_dir, no, data, size) < 0) {
			printf("ERROR: PNGファイルを書き込めません: %s%s%05d.png\n", out_dir, PATH_DELIM, no);
			goto ERROR_EXIT;
		}

		free(data);
		data = NULL;
	}

	result = 0;

ERROR_EXIT:
	if (dirp != NULL) {
		closedir(dirp);
	}
	if (data != NULL) {
		free(data);
	}

	return result;
}

static int check_png_name(const char *filename)
{
	int i;
	const char *ptr = filename;

	for (i = 0; i < 5; i++) {
		if (!isdigit(*ptr++)) {
			return -1;
		}
	}

	if (*ptr++ != '.') {
		return -1;
	}

	if ((ptr = strstr(ptr, ".png")) == NULL) {
		return -1;
	}

	if (*(ptr + 4) != 0) {
		return -1;
	}

	return 0;
}

static int read_png_file(const char *dir, const char *filename, unsigned char **data, long *size)
{
	int result = -1;
	FILE *fp = NULL;
	char *read_path = NULL;

	*data = NULL;
	*size = 0;

#ifdef WIN32
	{
		long alloc_size = strlen(dir) + strlen(PATH_DELIM) + strlen(filename) + 1;
		if ((read_path = (char *)malloc(alloc_size)) == NULL) {
			printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
			goto ERROR_EXIT;
		}
		strcat(strcat(strcpy(read_path, dir), PATH_DELIM), filename);
	}
#else
	if (asprintf(&read_path, "%s%s%s", dir, PATH_DELIM, filename) < 0) {
		printf("ERROR: asprintf(%s%s%s)\n", dir, PATH_DELIM, filename);
		goto ERROR_EXIT;
	}
#endif

	printf("read_path = %s\n", read_path);
	fflush(stdout);

	if ((fp = fopen(read_path, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_END: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	*size = ftell(fp);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_SET: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	if ((*data = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", *size);
		goto ERROR_EXIT;
	}

	if (fread(*data, *size, 1, fp) != 1) {
		printf("ERROR: %d: ファイル読み込みエラー: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	if (memcmp(*data, png_sig, 8) != 0) {
		printf("ERROR: PNGファイルではありません: %s\n", read_path);
		goto ERROR_EXIT;
	} else if (*size < 0x21) {
		printf("ERROR: ファイルサイズが小さすぎます: %s\n", read_path);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (read_path != NULL) {
		free(read_path);
	}
	if (result < 0) {
		if (*data != NULL) {
			free(*data);
			*data = NULL;
		}
		*size = 0;
	}

	return result;
}

static int write_png_file(const char *dir, int no, const unsigned char *data, long size)
{
	int result = -1;
	FILE *fp = NULL;
	char *write_path = NULL;
	struct stat st;

	if (stat(dir, &st) == 0) {
		if (!S_ISDIR(st.st_mode)) {
			printf("ERROR: ディレクトリではありません: %s\n", dir);
			goto ERROR_EXIT;
		}
	} else {
#ifdef WIN32
		if (mkdir(dir) < 0)
#else
		if (mkdir(dir, 0700) < 0)
#endif
		{
			printf("ERROR: %d: ディレクトリを作成できません: %s\n", errno, dir);
			goto ERROR_EXIT;
		}
	}

#ifdef WIN32
	{
		long alloc_size = strlen(dir) + strlen(PATH_DELIM) + 5 + 1 + 3 + 1;
		if ((write_path = (char *)malloc(alloc_size)) == NULL) {
			printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
			goto ERROR_EXIT;
		}
		snprintf(write_path, alloc_size, "%s%s%05d.png", dir, PATH_DELIM, no);
	}
#else
	if (asprintf(&write_path, "%s%s%05d.png", dir, PATH_DELIM, no) < 0) {
		printf("ERROR: asprintf(%s%s%05d.png)\n", dir, PATH_DELIM, no);
		goto ERROR_EXIT;
	}
#endif

	if ((fp = fopen(write_path, "wb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, write_path);
		goto ERROR_EXIT;
	}

	if (fwrite(data, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, write_path);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (write_path != NULL) {
		free(write_path);
	}

	return result;
}

static void mask_width_and_height(unsigned char *data, long size, const unsigned char *key)
{
	int i;

	for (i = 0; i < 8; i++) {
		printf("%02X ", (int)data[0x10 + i]);
	}
	printf("\n");

	for (i = 0; i < 8; i++) {
		data[0x10 + i] ^= (unsigned char)tolower(key[i]);
	}

	for (i = 0; i < 8; i++) {
		printf("%02X ", (int)data[0x10 + i]);
	}
	printf("\n");
}

static void make_png_key(unsigned char *key, const unsigned char *digest)
{
	int i;
	int value;

	for (i = 0; i < 4; i++) {
		value = (digest[i] >> 4) & 0xF;
		if (value < 10) {
			*key = '0' + value;
		} else {
			*key = 'a' + value - 10;
		}
		key++;

		value = digest[i] & 0xF;
		if (value < 10) {
			*key = '0' + value;
		} else {
			*key = 'a' + value - 10;
		}
		key++;
	}
}

/* EOF */
