#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


#include "utils.h"
#include "file_utils.h"
#include "font_info_sub.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static void usage(void);
static int modify_font_info(const char *in_file, const char *out_file, struct font_table_t *font_table);
static char *replace_text(const char *text, const char *start, const char *end, const char *new_text);


int main(int argc, char **argv)
{
	int num_cmd_params = 3;
	const char *xml_file = NULL;
	const char *font_file = NULL;
	const char *out_file = NULL;

	struct font_table_t font_table;

	memset(&font_table, 0, sizeof(font_table));

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 3) {
				xml_file = *argv;
			} else if (num_cmd_params == 2) {
				font_file = *argv;
			} else {
				out_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (xml_file == NULL || font_file == NULL || out_file == NULL) {
		usage();
		return 1;
	}

	if (read_font_info(&font_table, font_file) < 0) {
		printf("ERROR: フォント情報を読み込めません\n");
		return 1;
	}

	if (modify_font_info(xml_file, out_file, &font_table) < 0) {
		printf("ERROR: フォント情報を変更できません\n");
		return 1;
	}

	free_font_info(&font_table);

	return 0;
}

static void usage(void)
{
	printf("使用方法: modify_font [オプション] XMLファイル フォント情報ファイル 出力ファイル\n");
	printf("機    能: XMLファイルのフォント情報を変更します\n");
	printf("オプション:\n");
	printf("    -?|--help               使用方法を表示して終了します\n");
}

static int modify_font_info(const char *in_file, const char *out_file, struct font_table_t *font_table)
{
	int result = -1;
	FILE *in_fp = NULL;
	FILE *out_fp = NULL;
	int line = 0;
	char *ptr;
	char *mod_buf = NULL;
	const struct font_t141_t *font;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	if ((in_fp = fopen(in_file, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, in_file);
		goto ERROR_EXIT;
	}

	if ((out_fp = fopen(out_file, "wb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, out_file);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
			printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;

		if ((ptr = strstr(buffer, "<T141 ")) != NULL) {
			/* フォント */
			int offset;
			char *start;
			char *end;
			int len;
			int font_no = -1;

			/* フォントID */
			if ((start = strstr(ptr, " T13=\"")) == NULL) {
				printf("ERROR: フォントIDが見つかりません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			start += 6;

			if ((font_no = strtol(start, &end, 10)) < 0 || end == NULL || *end != '\"' || start == end) {
				printf("ERROR: 不正なフォントIDです: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}

			if ((font = find_font(font_table, font_no)) == NULL) {
				/* 対象外 */
				if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
					printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
					goto ERROR_EXIT;
				}
				continue;
			}

			if ((mod_buf = strdup(buffer)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			ptr += 5;
			offset = ptr - buffer;
			ptr = mod_buf + offset;

			/* フォント名 */
			if ((start = strstr(ptr, " T12=\"")) == NULL) {
				printf("ERROR: フォント名が見つかりません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			start += 6;

			if ((end = strchr(start, '\"')) == NULL) {
				printf("ERROR: 文字列終端の\"がありません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			len = end - start;

			if (font->t12 != NULL && (strlen(font->t12) != len || strncmp(font->t12, start, len) != 0)) {
				printf("WARNING: フォント名が違います: %s: %d行目\n", in_file, line);
				printf("\tXML: %.*s\n", len, start);
				printf("\t[%d]: %s\n", font_no, font->t12);
			}

			/* 文字間隔 */
			if ((start = strstr(ptr, " T703=\"")) == NULL) {
				printf("ERROR: 文字間隔が見つかりません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			start += 7;

			if ((end = strchr(start, '\"')) == NULL) {
				printf("ERROR: 文字列終端の\"がありません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			len = end - start;

			if (font->t703 != NULL) {
				char *new_buf;

				if ((new_buf = replace_text(mod_buf, start, end, font->t703)) == NULL) {
					goto ERROR_EXIT;
				}

				free(mod_buf);
				mod_buf = new_buf;
				ptr = mod_buf + offset;
			}

			/* 文字セット */
			if ((start = strstr(ptr, " T249=\"")) == NULL) {
				printf("ERROR: 文字セットが見つかりません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			start += 7;

			if ((end = strchr(start, '\"')) == NULL) {
				printf("ERROR: 文字列終端の\"がありません: %s: %d行目\n", in_file, line);
				goto ERROR_EXIT;
			}
			len = end - start;

			{
				char *new_buf;

				if ((new_buf = replace_text(mod_buf, start, end, font->t249)) == NULL) {
					goto ERROR_EXIT;
				}

				free(mod_buf);
				mod_buf = new_buf;
				ptr = mod_buf + offset;
			}

			if (fwrite(mod_buf, strlen(mod_buf), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
			free(mod_buf);
			mod_buf = NULL;

			while (1) {
				if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
					printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
					goto ERROR_EXIT;
				} else if (read_status == 0) {
					break;
				}

				line++;

				if ((ptr = strstr(buffer, "</T141>")) != NULL) {
					if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
						printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
						goto ERROR_EXIT;
					}
					break;
				} else if ((ptr = strstr(buffer, "<T248 ")) != NULL) {
					ptr += 5;

					/* パス */
					if ((start = strstr(buffer, " T7=\"")) == NULL) {
						printf("ERROR: パスが見つかりません: %s: %d行目\n", in_file, line);
						goto ERROR_EXIT;
					}
					start += 5;

					if ((end = strchr(start, '\"')) == NULL) {
						printf("ERROR: 文字列終端の\"がありません: %s: %d行目\n", in_file, line);
						goto ERROR_EXIT;
					}
					len = end - start;

					if (font->t248.t7 != NULL && (strlen(font->t248.t7) != len || strncmp(font->t248.t7, start, len) != 0)) {
						printf("WARNING: パスが違います: %s: %d行目\n", in_file, line);
						printf("\tXML: %.*s\n", len, start);
						printf("\t[%d]: %s\n", font_no, font->t248.t7);
					}

					if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
						printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
						goto ERROR_EXIT;
					}
				} else if ((ptr = strstr(buffer, "<T143>")) != NULL) {
					int found_t22 = 0;

					if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
						printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
						goto ERROR_EXIT;
					}

					while (1) {
						if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
							printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
							goto ERROR_EXIT;
						} else if (read_status == 0) {
							break;
						}

						if ((ptr = strstr(buffer, "</T143>")) != NULL) {
							if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
								printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
								goto ERROR_EXIT;
							}
							break;
						} else if ((ptr = strstr(buffer, "<T22 ")) != NULL) {
							int i;
							char *crlf;

							/* 最初に<T22を見つけたときに全て出力する */
							if (found_t22) {
								continue;
							}
							found_t22 = 1;

							len = ptr - buffer;
							if ((crlf = strpbrk(ptr, "\r\n")) == NULL) {
								/* ここには来ないはず */
								crlf = "\r\n";
							}
							for (i = 0; i < font->t143->size; i++) {
								const struct font_t22_t *t22 = &(font->t143->t22[i]);

								if (fwrite(buffer, len, 1, out_fp) != 1 ||
									fprintf(out_fp, "<T22 T3=\"%d\" T4=\"%d\" T5=\"%d\" T6=\"%d\"/>%s", t22->t3, t22->t4, t22->t3 + t22->t5 - 1, t22->t4 + t22->t6 - 1, crlf) < 0) {
									printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
									goto ERROR_EXIT;
								}
							}
						} else {
							if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
								printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
								goto ERROR_EXIT;
							}
						}
					}
				} else {
					if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
						printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
						goto ERROR_EXIT;
					}
				}
			}
		} else {
			/* 書き換え対象以外 */
			if (fwrite(buffer, strlen(buffer), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
		}
	}

	result = 0;

ERROR_EXIT:
	if (out_fp != NULL) {
		fclose(out_fp);
	}
	if (in_fp != NULL) {
		fclose(in_fp);
	}
	if (mod_buf != NULL) {
		free(mod_buf);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static char *replace_text(const char *text, const char *start, const char *end, const char *new_text)
{
	int len = end - start;
	int left_len = start - text;
	size_t alloc_size = strlen(text) - len + strlen(new_text) + 1;
	char *buffer;

	if ((buffer = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return NULL;
	}
	strncpy(buffer, text, left_len);
	strcat(strcpy(buffer + left_len, new_text), end);

	return buffer;
}

/* EOF */
