#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


#include "utils.h"
#include "file_utils.h"
#include "font_info_sub.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static void usage(void);
static void free_rect_file_table(struct rect_file_table_t *rect_file_table);
static int read_rect_file_list(struct rect_file_table_t *rect_file_table, const char *filename);
static int parse_file_info(char *src, int *index, char **rect_file);
static int read_rect_files(struct rect_file_table_t *rect_file_table);
static int modify_font_info(struct font_table_t *font_table, struct rect_file_table_t *rect_file_table);
static int write_font_info_main(const struct font_table_t *font_table, const char *filename);
static int write_font_info(FILE *fp, const struct font_t141_t *font);
static int print_safe_text(const char *text);
static int write_safe_text(FILE *fp, const char *text);


int main(int argc, char **argv)
{
	int num_cmd_params = 3;
	const char *font_file = NULL;
	const char *list_file = NULL;
	const char *out_file = NULL;

	struct font_table_t font_table;
	struct rect_file_table_t rect_file_table;

	memset(&font_table, 0, sizeof(font_table));
	memset(&rect_file_table, 0, sizeof(rect_file_table));

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 3) {
				font_file = *argv;
			} else if (num_cmd_params == 2) {
				list_file = *argv;
			} else {
				out_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (font_file == NULL || list_file == NULL || out_file == NULL) {
		usage();
		return 1;
	}

	if (read_font_info(&font_table, font_file) < 0) {
		printf("ERROR: フォント情報を読み込めません\n");
		return 1;
	}

	if (read_rect_file_list(&rect_file_table, list_file) < 0) {
		printf("ERROR: 座標ファイル一覧を読み込めません\n");
		return 1;
	}

	if (read_rect_files(&rect_file_table) < 0) {
		printf("ERROR: 座標ファイルを読み込めません\n");
		return 1;
	}

	if (modify_font_info(&font_table, &rect_file_table) < 0) {
		printf("ERROR: フォント情報を変更できません\n");
		return 1;
	}

	if (write_font_info_main(&font_table, out_file) < 0) {
		printf("ERROR: フォント情報を書き込めません\n");
		return 1;
	}

	free_rect_file_table(&rect_file_table);
	free_font_info(&font_table);

	return 0;
}

static void usage(void)
{
	printf("使用方法: modify_font_info [オプション] フォント情報ファイル 座標ファイル一覧\n");
	printf("                                                                   出力ファイル\n");
	printf("機    能: 座標ファイル一覧で指定された座標ファイルの内容でフォント情報を変更し\n");
	printf("          ます\n");
	printf("オプション:\n");
	printf("    -?|--help               使用方法を表示して終了します\n");
}

static void free_rect_file_table(struct rect_file_table_t *rect_file_table)
{
	if (rect_file_table->table != NULL) {
		int i;

		for (i = 0; i < rect_file_table->size; i++) {
			if (rect_file_table->table[i].rect_file != NULL) {
				free(rect_file_table->table[i].rect_file);
			}
			free_font_charset(&(rect_file_table->table[i].charset));
		}

		free(rect_file_table->table);
	}

	memset(rect_file_table, 0, sizeof(*rect_file_table));
}

static int read_rect_file_list(struct rect_file_table_t *rect_file_table, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	char *src;
	int line = 0;

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	if ((fp = fopen(filename, "rt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	while (1) {
		int index = -1;
		char *rect_file = NULL;
		int res;

		if ((read_status = read_next_line(&buffer, &buffer_size, fp)) < 0) {
			printf("ERROR: 読み込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;
		src = buffer;

		if (line == 1) {
			src = skip_bom(src);
		}

		if (*src == '#' || *src == ';') {
			continue;
		}

		if (parse_file_info(src, &index, &rect_file) < 0) {
			printf("ERROR: 不正なファイル定義です: %s: %d行目\n", filename, line);
			goto ERROR_EXIT;
		}

		if (index < 0 || rect_file == NULL) {
			continue;
		}

		if ((res = add_rect_file(rect_file_table, index, rect_file)) < 0) {
			goto ERROR_EXIT;
		} else if (res == 0) {
			printf("WARNING: 登録済のインデックスです: %s: %d行目: index=%d, file=%s\n", filename, line, index, rect_file);
		}
	}

	if (rect_file_table->size == 0) {
		printf("ERROR: 座標ファイル情報がありません: %s\n", filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static int parse_file_info(char *src, int *index, char **rect_file)
{
	const char *separater = "\t\r\n";
	char *token;
	char *endptr;

	*index = -1;
	*rect_file = NULL;

	if ((token = strtok(src, separater)) == NULL) {
		return 0;
	}

	*index = strtol(token, &endptr, 10);
	if (endptr == NULL || *endptr != 0 || *index < 0) {
		printf("ERROR: 不正なインデックスです: %s\n", token);
		return -1;
	}

	if ((token = strtok(NULL, separater)) == NULL) {
		printf("ERROR: 座標ファイル名がありません\n");
		return -1;
	}

	if ((*rect_file = strdup(token)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}

	if (strtok(NULL, separater) != NULL) {
		return -1;
	}

	return 0;
}

static int read_rect_files(struct rect_file_table_t *rect_file_table)
{
	int i;

	for (i = 0; i < rect_file_table->size; i++) {
		if (read_char_rect(&(rect_file_table->table[i].charset), rect_file_table->table[i].rect_file) < 0) {
			return -1;
		}
	}

	return 0;
}

static int modify_font_info(struct font_table_t *font_table, struct rect_file_table_t *rect_file_table)
{
	int result = -1;
	int i;

	for (i = 0; i < font_table->size; i++) {
		struct font_t141_t *t141 = &(font_table->t141[i]);
		struct font_t143_t *t143 = t141->t143;
		int font_no = t141->t13;
		const char *path = t141->t248.t7;
		int index = -1;
		const struct rect_file_info_t *info;

		/* pathの最後の5桁数字がindexと同じ座標情報を探す */
		if (t141->ref_t13 < 0) {
			if ((index = get_path_index(path)) < 0) {
				printf("ERROR: 不正なpathです: フォント%d, path=", font_no);
				print_safe_text(path);
				printf("\n");
				goto ERROR_EXIT;
			}
		} else {
			/* 別フォントを参照しているとき */
			const struct font_t141_t *ref_t141 = NULL;
			const char *ref_path = NULL;
			int ref_index = -1;

			if (path == NULL || *path == '\0') {
				/* path指定なし ... 参照先の画像を使用する */
				/* 変更は参照先で行うのでこのフォントでは何もしない */
				continue;
			}

			if ((index = get_path_index(path)) < 0) {
				printf("ERROR: 不正なpathです: フォント%d, path=", ref_t141->t13);
				print_safe_text(path);
				printf("\n");
				goto ERROR_EXIT;
			}

			if ((ref_t141 = find_font(font_table, t141->ref_t13)) == NULL) {
				printf("ERROR: フォント%d: charset=[%d]で参照しているフォントが見つかりません\n", t141->t13, t141->ref_t13);
				goto ERROR_EXIT;
			}

			ref_path = ref_t141->t248.t7;
			if ((ref_index = get_path_index(ref_path)) < 0) {
				printf("ERROR: 不正なpathです: フォント%d, path=", ref_t141->t13);
				print_safe_text(ref_path);
				printf("\n");
				goto ERROR_EXIT;
			}

			if (index == ref_index) {
				/* 参照先と同じ画像を指定しているとき */
				/* 変更は参照先で行うのでこのフォントでは何もしない */
				continue;
			}

			/*
			 * 参照元と参照先で別の画像を指定しているとき
			 * 参照元画像の座標情報を変更するとcharset=[*]が解除される
			 *
			 * 別画像を指定していても、参照元画像の座標情報が無ければ
			 * charset=[*]は維持され、参照先の座標情報が使用される
			 */
		}

		if ((info = find_rect_file_info(rect_file_table, index)) == NULL) {
			/* 更新定義ファイルに未登録 */
#if 0
			printf("DEBUG: [%d] index not found: #%05d#\n", t141->t13, index);
#endif
			continue;
		}

		if (info->charset.used == 0) {
			/* 変更なし */
			continue;
		}

		if (t141->ref_t13 < 0) {
			/* char=...で定義したフォント → 定義内容を入れ替える */
			/* 旧定義の開放は新定義に入れ替え後、参照している全てのフォントのポインタを変更してから行う */
			t141->t143 = NULL;
		} else {
			/* charset=[*]で参照しているフォント → 新しい定義を作成する */
			t141->t143->link_count--;
			t141->t143 = NULL;
			t143 = NULL;			/* 参照元フォント側のポインタは維持されているのでここでNULLにしても大丈夫 */
			t141->ref_t13 = -1;		/* charset=[*]解除 */
		}

		if (t141->t249 != NULL) {
			free(t141->t249);
		}
		t141->t249 = NULL;

		if (convert_font_charset(t141, &(info->charset)) < 0) {
			goto ERROR_EXIT;
		}

		if (t143 != NULL) {
			int j;
			t141->t143->link_count = t143->link_count;

			/* 同じ定義を参照していたフォントの参照先を更新 */
			for (j = 0; j < font_table->size; j++) {
				if (font_table->t141[j].t143 == t143) {
					font_table->t141[j].t143 = t141->t143;
				}
			}

			if (t143->t22 != NULL) {
				free(t143->t22);
			}
			free(t143);
		}
	}

	result = 0;


ERROR_EXIT:

	return result;
}

static int write_font_info_main(const struct font_table_t *font_table, const char *filename)
{
	int result = -1;
	FILE *fp = NULL;
	int i;

	if ((fp = fopen(filename, "wt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	for (i = 0; i < font_table->size; i++) {
		if (write_font_info(fp, &(font_table->t141[i])) < 0) {
			printf("ERROR: 書き込みエラー: %s\n", filename);
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int write_font_info(FILE *fp, const struct font_t141_t *font)
{
	int i;
	const char *src = font->t249;

	if (fprintf(fp, "[%d]\n", font->t13) < 0) {
		return -1;
	}

	if (fprintf(fp, "name=") < 0 ||
		write_safe_text(fp, font->t12) < 0 ||
		fprintf(fp, "\npath=") < 0 ||
		write_safe_text(fp, font->t248.t7) < 0 ||
		fprintf(fp, "\n") < 0) {
		return -1;
	}

	if (0 <= font->ref_t13) {
		if (fprintf(fp, "charset=[%d]\n", font->ref_t13) < 0) {
			return -1;
		}
	} else {
		if (fprintf(fp, "spacing=") < 0 ||
			write_safe_text(fp, font->t703) < 0 ||
			fprintf(fp, "\n") < 0) {
			return -1;
		}

		for (i = 0; i < font->t143->size; i++) {
			const struct font_t22_t *t22 = &(font->t143->t22[i]);
			unsigned long code;
			int len;

			if ((len = get_next_utf8_char(&code, src, 1)) < 0) {
				printf("ERROR: フォント%d: 文字セットに不正な文字が含まれています\n", font->t13);
				return -1;
			}
			if (len == 0) {
				printf("ERROR: フォント%d: 文字セットが不足しています\n", font->t13);
				return -1;
			}

			if (fprintf(fp, "char=") < 0) {
				return -1;
			}

			if (code <= 0x7F) {
				if (fprintf(fp, "%c", (char)(code & 0x7F)) < 0) {
					return -1;
				}
			} else {
				if (fwrite(src, len, 1, fp) != 1) {
					return -1;
				}
			}

			if (fprintf(fp, "\t%d\t%d\t%d\t%d\n", t22->t4, t22->t3, t22->t4 + t22->t6 - 1, t22->t3 + t22->t5 - 1) < 0) {
				return -1;
			}

			src += len;
		}

		if (*src != 0) {
			printf("WARNING: フォント%d: 座標データが不足しています\n", font->t13);
			printf("\tname=");
			print_safe_text(font->t12);
			printf("\n\tpath=");
			print_safe_text(font->t248.t7);
			printf("\n");
		}
	}

	if (fprintf(fp, "\n") < 0) {
		return -1;
	}

	return 0;
}

static int print_safe_text(const char *text)
{
	if (text != NULL && *text != 0) {
		if (printf("%s", text) < 0) {
			return -1;
		}
	}

	return 0;
}

static int write_safe_text(FILE *fp, const char *text)
{
	if (text != NULL && *text != 0) {
		if (fprintf(fp, "%s", text) < 0) {
			return -1;
		}
	}

	return 0;
}

/* EOF */
