#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "utils.h"
#include "file_utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static void usage(void);
static int pickup_text(const char *in_file, const char *out_file, int lang_no, int output_speech_path);
static int quote_text(char **des, const char *src);


int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *xml_file = NULL;
	const char *text_file = NULL;

	int lang_no = 1;
	int output_speech_path = 0;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-l") == 0 || strcmp(*argv, "--lang") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("ERROR: 言語番号がありません\n");
					return 1;
				}
				lang_no = strtoul(*argv, &endptr, 10);
				if (endptr == NULL || *endptr != 0 || lang_no < 0) {
					printf("ERROR: 不正な言語番号です: %s\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--speech-path") == 0) {
				output_speech_path = 1;
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				xml_file = *argv;
			} else {
				text_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (xml_file == NULL || text_file == NULL) {
		usage();
		return 1;
	}

	if (pickup_text(xml_file, text_file, lang_no, output_speech_path) < 0) {
		printf("ERROR: テキスト抽出できません\n");
		return 1;
	}

	return 0;
}

static void usage(void)
{
	printf("使用方法: pickup_text [オプション] XMLファイル 出力ファイル\n");
	printf("機    能: XMLファイルのテキストをタブ区切りで保存します\n");
	printf("オプション:\n");
	printf("    -l|--lang 言語番号  テキストの言語番号を指定します (デフォルト:1)\n");
	printf("       --speech-path    音声ファイルのパスを追加出力します\n");
	printf("    -?|--help           使用方法を表示して終了します\n");
}

static int pickup_text(const char *in_file, const char *out_file, int lang_no, int output_speech_path)
{
	int result = -1;
	FILE *in_fp = NULL;
	FILE *out_fp = NULL;
	long line = 0;
	long line2 = 0;
	long id;
	char id_str[32];
	char *ptr;
	char *ptr2;
	char *text;
	int text_len;
	char *text2 = NULL;
	char *speech;
	int speech_len;
	const char *header = output_speech_path ? "id\t原文\t翻訳\t翻訳\t音声ファイル\r\n" : "id\t原文\t翻訳\t翻訳\r\n";

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;
	char *buffer2 = NULL;
	size_t buffer2_size = 0;

	char t13_pattern[32];

	snprintf(t13_pattern, sizeof(t13_pattern), " T13=\"%d\"", lang_no);

	if ((in_fp = fopen(in_file, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, in_file);
		goto ERROR_EXIT;
	}

	if ((out_fp = fopen(out_file, "wb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, out_file);
		goto ERROR_EXIT;
	}

	if (fwrite(header, strlen(header), 1, out_fp) != 1) {
		printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, in_fp)) < 0) {
			printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		line++;

		if ((ptr = strstr(buffer, "<T255 ")) == NULL) {
			/* テキストでない */
			continue;
		}

		/* <T258 ... />の行を追加読み込み */
		line2 = line;
		ptr2 = NULL;
		while (1) {
			if ((read_status = read_next_line(&buffer2, &buffer2_size, in_fp)) < 0) {
				line = line2;
				printf("ERROR: ファイル読み込みエラー: %s\n", in_file);
				goto ERROR_EXIT;
			} else if (read_status == 0) {
				break;
			}
			line2++;

			if ((ptr2 = strstr(buffer2, "<T258 ")) != NULL) {
				break;
			}
			if (strstr(buffer2, "</T255>") != NULL) {
				break;
			}
		}
		if (ptr2 == NULL) {
			printf("ERROR: <T258 ... />が見つかりません: %s: line=%ld\n", in_file, line2);
			goto ERROR_EXIT;
		}

/*		if (strstr(ptr2 + 5, " T13=\"1\"") == NULL) {*/
		if (strstr(ptr2 + 5, t13_pattern) == NULL) {
			/* 対象外言語のテキスト */
			line = line2;
			continue;
		}

		if ((ptr = strstr(ptr + 5, " T13=\"")) == NULL) {
			printf("ERROR: テキストIDが見つかりません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		id = strtoul(ptr + 6, &ptr, 10);
		if (*ptr != '\"') {
			printf("ERROR: テキストIDの書式が違います: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		if (output_speech_path) {
			if ((ptr = strstr(ptr + 1, " T257=\"")) == NULL) {
				printf("WARNING: 音声ファイルが見つかりません: %s(%ld)\n", in_file, line);
			} else {
				speech = ptr + 7;

				if ((ptr = strstr(speech, "\"")) == NULL) {
					printf("ERROR: 音声ファイルが\"で終っていません: %s(%ld)\n", in_file, line);
					goto ERROR_EXIT;
				}

				if ((speech_len = ptr - speech) == 0) {
					speech = NULL;
				} else if (unescape_text(speech, ptr) < 0) {
					printf("ERROR: 不正な文字が使用されています: %s(%ld)\n", in_file, line);
					goto ERROR_EXIT;
				}
			}
		}

		if ((ptr = strstr(ptr + 1, " T256=\"")) == NULL) {
			printf("ERROR: テキストが見つかりません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		text = ptr + 7;

		if ((ptr = strstr(text, "\"")) == NULL) {
			printf("ERROR: テキストが\"で終っていません: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		if ((text_len = ptr - text) == 0) {
			line = line2;
			continue;
		}

		if (unescape_text(text, ptr) < 0) {
			printf("ERROR: 不正な文字が使用されています: %s(%ld)\n", in_file, line);
			goto ERROR_EXIT;
		}

		if (quote_text(&text2, text) < 0) {
			printf("ERROR: 文字列変換できません\n");
			goto ERROR_EXIT;
		}
		if (text2 != NULL) {
			text = text2;
		}

		snprintf(id_str, sizeof(id_str), "%ld\t", id);

		if (fwrite(id_str, strlen(id_str), 1, out_fp) != 1) {
			printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		if (fwrite(text, strlen(text), 1, out_fp) != 1) {
			printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
			goto ERROR_EXIT;
		}

		if (output_speech_path && speech != NULL) {
			if (fwrite("\t\t\t", 3, 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
			if (fwrite(speech, strlen(speech), 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
			if (fwrite("\r\n", 2, 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
		} else {
			if (fwrite("\t\t\r\n", 4, 1, out_fp) != 1) {
				printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, out_file);
				goto ERROR_EXIT;
			}
		}

		if (text2 != NULL) {
			free(text2);
			text2 = NULL;
		}

		line = line2;
	}

	result = 0;

ERROR_EXIT:
	if (in_fp != NULL) {
		fclose(in_fp);
	}
	if (out_fp != NULL) {
		fclose(out_fp);
	}
	if (text2 != NULL) {
		free(text2);
	}
	if (buffer != NULL) {
		free(buffer);
	}
	if (buffer2 != NULL) {
		free(buffer2);
	}

	return result;
}

static int quote_text(char **des, const char *src)
{
	const char *ptr;
	int count = 0;
	long alloc_size;
	char *des2;

	*des = NULL;

	for (ptr = src; (ptr = strchr(ptr, '\"')) != NULL; ptr++) {
		count++;
	}

	if (count == 0) {
		return 0;
	}

	alloc_size = strlen(src) + count + 2 + 1;
	if ((*des = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
		return -1;
	}
	des2 = *des;

	*des2++ = '\"';

	while (*src != 0) {
		if (*src == '\"') {
			*des2++ = *src;
		}
		*des2++ = *src++;
	}

	*des2++ = '\"';
	*des2 = 0;

	return 1;
}

/* EOF */
