/*
 * PNG画像のマスク処理はVISExtから移植しました。
 * VISExt 1.2.5174 Copyright (C) 2010-2014, Deniz Oezmen (http://oezmen.eu/)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#ifdef WIN32
#include "md5.h"
#define MD5_CTX md5_state_t
#define MD5Init(_p) md5_init(_p)
#define MD5Update(_p,_d,_n) md5_append(_p,(const md5_byte_t *)_d,_n)
#define MD5Final(_d,_p) md5_finish(_p,_d)
#else
#include <md5.h>
#endif
#include <zlib.h>

#include "file_utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


#ifdef WIN32
#define PATH_DELIM	"\\"
#else
#define PATH_DELIM	"/"
#endif


static const unsigned char png_sig[8] = {
	0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A
};

struct file_table_t {
	int size;
	struct file_table_entry_t {
		int no;
		char *filename;
	} *table;
};


static void usage(void);
static int read_xml_file(struct file_table_t *table, const char *xml_filename, char vis_type, int vis_no);
static void init_file_table(struct file_table_t *table);
static int add_file_table(struct file_table_t *table, int no, const char *filename);
static void free_file_table(struct file_table_t *table);
static int unmask_png_main(struct file_table_t *table, const char *in_dir, const char *out_dir);
static int unmask_png(const char *in_dir, const char *out_dir, int no, const char *orig_filename);
static int read_png_file(const char *dir, const char *filename, unsigned char **data, long *size);
static int write_png_file(int no, const char *dir, const char *filename, const unsigned char *data, long size);
static unsigned long get_unsigned_long(const unsigned char *data);
static int unmask_width_and_height(unsigned char *data, long size, const unsigned char *key, unsigned long target_crc);
static void make_png_key(unsigned char *key, const unsigned char *digest);


int main(int argc, char **argv)
{
	int result = 1;
	int num_cmd_params = 3;
	const char *xml_file = NULL;
	const char *in_dir = NULL;
	const char *out_dir = NULL;
	struct file_table_t table;
	char vis_type = 'v';
	int vis_no = -1;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "-t") == 0 || strcmp(*argv, "--type") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: アーカイブ種別がありません\n");
					return 1;
				}
				if (strlen(*argv) != 1 || (**argv != 'c' && **argv != 's' && **argv != 'v')) {
					printf("ERROR: 不正なアーカイブ種別です: %s\n", *argv);
					return 1;
				}
				vis_type = **argv;
			} else if (strcmp(*argv, "-n") == 0 || strcmp(*argv, "--no") == 0) {
				char *endptr;
				if (*++argv == NULL) {
					printf("ERROR: アーカイブ番号がありません\n");
					return 1;
				}
				vis_no = strtol(*argv, &endptr, 10);
				if (endptr == NULL || *endptr != 0 || vis_no < 0 || 999 < vis_no) {
					printf("ERROR: 不正なアーカイブ番号です: %s\n", *argv);
					return 1;
				}
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
			if (num_cmd_params == 3) {
				xml_file = *argv;
			} else if (num_cmd_params == 2) {
				in_dir = *argv;
			} else {
				out_dir = *argv;
			}
			num_cmd_params--;
		}
	}

	if (vis_type == 'v') {
		vis_no = -1;
	} else if (vis_no < 0) {
		printf("ERROR: アーカイブ番号が指定されていません\n");
		return 1;
	}

	if (in_dir == NULL || out_dir == NULL || xml_file == NULL) {
		usage();
		return 1;
	}

	init_file_table(&table);

	if (read_xml_file(&table, xml_file, vis_type, vis_no) < 0) {
		printf("ERROR: XMLファイルを読み込めません\n");
		goto ERROR_EXIT;
	}

	if (unmask_png_main(&table, in_dir, out_dir) < 0) {
		printf("ERROR: 変換エラーが発生しました\n");
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	free_file_table(&table);

	return result;
}

static void usage(void)
{
	printf("使用方法: unmask_png [オプション] XMLファイル 入力DIR 出力DIR\n");
	printf("機    能: 画像ファイルのマスクを解除します\n");
	printf("オプション:\n");
	printf("    -t|--type c|s|v     画像が格納されていたファイルの種類\n");
	printf("                            c   character.vc???\n");
	printf("                            s   scene.vs???\n");
	printf("                            v   data.vis\n");
	printf("    -n|--no 番号        画像が格納されていたファイルの番号(ファイル名末尾の数値)\n");
	printf("                        --typeオプションでvを指定したときは無視します\n");
	printf("    -?|--help           使用方法を表示して終了します\n");
}

static int read_xml_file(struct file_table_t *table, const char *xml_filename, char vis_type, int vis_no)
{
	int result = -1;
	FILE *fp = NULL;
	char *startptr;
	char *endptr;
	char *ptr;
	char *basename = NULL;
	int basename_len = 0;
	char *orig_filename = NULL;
	int no;
	char file_type_str[32];

	int read_status;
	char *buffer = NULL;
	size_t buffer_size = 0;

	if (vis_no < 0) {
		snprintf(file_type_str, sizeof(file_type_str), ".png#%c#-01#", vis_type);
	} else {
		snprintf(file_type_str, sizeof(file_type_str), ".png#%c#%03d#", vis_type, vis_no);
	}

	if ((fp = fopen(xml_filename, "rt")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, xml_filename);
		goto ERROR_EXIT;
	}

	while (1) {
		if ((read_status = read_next_line(&buffer, &buffer_size, fp)) < 0) {
			printf("ERROR: ファイル読み込みエラー: %s\n", xml_filename);
			goto ERROR_EXIT;
		} else if (read_status == 0) {
			break;
		}

		if ((startptr = strstr(buffer, "\"data\\")) == NULL) {
			continue;
		}
		startptr++;
		if ((endptr = strstr(startptr, file_type_str)) == NULL) {
			continue;
		}

		basename = startptr;
		for (ptr = startptr; ptr < endptr; ptr++) {
			if (*ptr == '\"') {
				basename = NULL;
				break;
			}
			if (*ptr == '\\') {
				basename = ptr + 1;
			}
		}
		if (basename == NULL) {
			continue;
		}
		basename_len = endptr + 4 - basename;

		ptr = endptr + 11;
		no = strtoul(ptr, &endptr, 10);
		if (endptr == NULL || endptr != ptr + 5) {
			continue;
		}

		if ((orig_filename = (char *)malloc(basename_len + 1)) == NULL) {
			printf("ERROR: memory exhausted: size=%d\n", basename_len + 1);
			goto ERROR_EXIT;
		}

		strncpy(orig_filename, basename, basename_len);
		orig_filename[basename_len] = 0;

		if (add_file_table(table, no, orig_filename) < 0) {
			printf("ERROR: ファイル名を登録できません: %05d, %s\n", no, orig_filename);
			goto ERROR_EXIT;
		}

		free(orig_filename);
		orig_filename = NULL;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (orig_filename != NULL) {
		free(orig_filename);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

static void init_file_table(struct file_table_t *table)
{
	table->size = 0;
	table->table = NULL;
}

static int add_file_table(struct file_table_t *table, int no, const char *filename)
{
	int i;
	int new_size;
	int alloc_size;

	for (i = 0; i < table->size; i++) {
		if (table->table[i].no < 0) {
			break;
		}

		if (table->table[i].no == no) {
			if (strcmp(table->table[i].filename, filename) != 0) {
				printf("WARNING: 同じ番号に別のファイル名が登録されています ... 登録しません: %05d, %s, %s\n", no, table->table[i].filename, filename);
			}
			return 0;
		}
	}

	if (table->size <= i) {
		new_size = table->size + 100;
		alloc_size = sizeof(struct file_table_entry_t) * new_size;
		if (table->size == 0) {
			table->table = (struct file_table_entry_t *)malloc(alloc_size);
		} else {
			table->table = (struct file_table_entry_t *)realloc(table->table, alloc_size);
		}
		if (table->table == NULL) {
			printf("ERROR: memory exhausted: size=%d\n", alloc_size);
			return -1;
		}

		for (i = table->size; i < new_size; i++) {
			table->table[i].no = -1;
			table->table[i].filename = NULL;
		}

		i = table->size;
		table->size = new_size;
	}

	if ((table->table[i].filename = strdup(filename)) == NULL) {
		printf("ERROR: memory exhausted: strdup(%s)\n", filename);
		return -1;
	}
	table->table[i].no = no;

	return 1;
}

static void free_file_table(struct file_table_t *table)
{
	free(table->table);
	table->table = NULL;
	table->size = 0;
}

static int unmask_png_main(struct file_table_t *table, const char *in_dir, const char *out_dir)
{
	int result = 0;
	int i;

	for (i = 0; i < table->size; i++) {
		int no = table->table[i].no;
		const char *orig_filename = table->table[i].filename;

		if (no < 0 || orig_filename == NULL) {
			continue;
		}

		printf("変換ファイル: %05d, %s\n", no, orig_filename);

		if (unmask_png(in_dir, out_dir, no, orig_filename) < 0) {
			printf("ERROR: ファイル変換できません: %05d, %s\n", no, orig_filename);
			result--;
		}

		printf("\n");
	}

	return result;
}

static int unmask_png(const char *in_dir, const char *out_dir, int no, const char *orig_filename)
{
	int result = -1;
	unsigned char *data = NULL;
	long size;
	unsigned long current_crc;
	unsigned long target_crc;
	MD5_CTX md5;
	unsigned char digest[16];
	unsigned char png_key[8];
	const char *basename;
	int basename_len;
	char filename[32];

	snprintf(filename, sizeof(filename), "%05d.png", no);

	if (read_png_file(in_dir, filename, &data, &size) < 0) {
		printf("ERROR: PNGファイルを読み込めません: %s%s%s\n", in_dir, PATH_DELIM, filename);
		goto ERROR_EXIT;
	}

	current_crc = crc32(0x00000000L, data + 0x0C, 0x11);
	printf("crc = %08lX\n", current_crc);
	target_crc = get_unsigned_long(data + 0x1D);
	printf("target = %08lX\n", target_crc);

	if (current_crc == target_crc) {
		printf("変換する必要はありません\n");
	} else {
		const char *ptr;

		basename = orig_filename;
		for (ptr = basename; *ptr != 0; ptr++) {
			if (*ptr == '\\') {
				basename = ptr + 1;
			}
		}

		if ((ptr = strrchr(basename, '.')) != NULL) {
			basename_len = ptr - basename;
		} else {
			basename_len = strlen(basename);
		}

		MD5Init(&md5);
		MD5Update(&md5, basename, basename_len);
		MD5Final(digest, &md5);

		printf("md5 = ");
		{ int i; for (i = 0; i < 16; i++) { printf("%02X", digest[i]); } }
		printf("\n");

		make_png_key(png_key, digest);

		if (unmask_width_and_height(data, size, png_key, target_crc) < 0) {
			printf("ERROR: 変換失敗\n");
			goto ERROR_EXIT;
		}
	}

	if (write_png_file(no, out_dir, orig_filename, data, size) < 0) {
		printf("ERROR: PNGファイルを書き込めません: %s%s%05d.%s\n", out_dir, PATH_DELIM, no, orig_filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (data != NULL) {
		free(data);
	}

	return result;
}

static int read_png_file(const char *dir, const char *filename, unsigned char **data, long *size)
{
	int result = -1;
	FILE *fp = NULL;
	char *read_path = NULL;

	*data = NULL;
	*size = 0;

#ifdef WIN32
	{
		long alloc_size = strlen(dir) + strlen(PATH_DELIM) + strlen(filename) + 1;
		if ((read_path = (char *)malloc(alloc_size)) == NULL) {
			printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
			goto ERROR_EXIT;
		}
		strcat(strcat(strcpy(read_path, dir), PATH_DELIM), filename);
	}
#else
	if (asprintf(&read_path, "%s%s%s", dir, PATH_DELIM, filename) < 0) {
		printf("ERROR: asprintf(%s%s%s)\n", dir, PATH_DELIM, filename);
		goto ERROR_EXIT;
	}
#endif

	if ((fp = fopen(read_path, "rb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_END: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	*size = ftell(fp);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("ERROR: %d: ファイルシークエラー: SEEK_SET: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	if ((*data = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", *size);
		goto ERROR_EXIT;
	}

	if (fread(*data, *size, 1, fp) != 1) {
		printf("ERROR: %d: ファイル読み込みエラー: %s\n", errno, read_path);
		goto ERROR_EXIT;
	}

	if (memcmp(*data, png_sig, 8) != 0) {
		printf("ERROR: PNGファイルではありません: %s\n", read_path);
		goto ERROR_EXIT;
	} else if (*size < 0x21) {
		printf("ERROR: ファイルサイズが小さすぎます: %s\n", read_path);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (read_path != NULL) {
		free(read_path);
	}
	if (result < 0) {
		if (*data != NULL) {
			free(*data);
			*data = NULL;
		}
		*size = 0;
	}

	return result;
}

static int write_png_file(int no, const char *dir, const char *filename, const unsigned char *data, long size)
{
	int result = -1;
	FILE *fp = NULL;
	char *write_path = NULL;
	struct stat st;

	if (stat(dir, &st) == 0) {
		if (!S_ISDIR(st.st_mode)) {
			printf("ERROR: ディレクトリではありません: %s\n", dir);
			goto ERROR_EXIT;
		}
	} else {
#ifdef WIN32
		if (mkdir(dir) < 0)
#else
		if (mkdir(dir, 0700) < 0)
#endif
		{
			printf("ERROR: %d: ディレクトリを作成できません: %s\n", errno, dir);
			goto ERROR_EXIT;
		}
	}

#ifdef WIN32
	{
		long alloc_size = strlen(dir) + strlen(PATH_DELIM) + 5 + 1 + strlen(filename) + 1;
		if ((write_path = (char *)malloc(alloc_size)) == NULL) {
			printf("ERROR: memory exhausted: size=%ld\n", alloc_size);
			goto ERROR_EXIT;
		}
		snprintf(write_path, alloc_size, "%s%s%05d.%s", dir, PATH_DELIM, no, filename);
	}
#else
	if (asprintf(&write_path, "%s%s%05d.%s", dir, PATH_DELIM, no, filename) < 0) {
		printf("ERROR: asprintf(%s%s%05d.%s)\n", dir, PATH_DELIM, no, filename);
		goto ERROR_EXIT;
	}
#endif

	if ((fp = fopen(write_path, "wb")) == NULL) {
		printf("ERROR: %d: オープンエラー: %s\n", errno, write_path);
		goto ERROR_EXIT;
	}

	if (fwrite(data, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイル書き込みエラー: %s\n", errno, write_path);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (write_path != NULL) {
		free(write_path);
	}

	return result;
}

static unsigned long get_unsigned_long(const unsigned char *data)
{
	unsigned long value = 0L;
	int i;

	for (i = 0; i < 4; i++) {
		value <<= 8;
		value |= *(data++);
	}

	return value;
}

static int unmask_width_and_height(unsigned char *data, long size, const unsigned char *key, unsigned long target_crc)
{
	unsigned char test_buf[0x11];
	int i;
	unsigned long crc;

	memcpy(test_buf, data + 0x0C, sizeof(test_buf));
	for (i = 0; i < sizeof(test_buf); i++) {
		printf("%02X ", (int)test_buf[i]);
	}
	printf("\n");

	for (i = 4; i < 12; i++) {
		test_buf[i] ^= (unsigned char)tolower(key[i - 4]);
	}

	for (i = 0; i < sizeof(test_buf); i++) {
		printf("%02X ", (int)test_buf[i]);
	}
	printf("\n");

	crc = crc32(0x00000000L, test_buf, sizeof(test_buf));
	printf("crc = %08lX\n", crc);

	if (crc != target_crc) {
		return -1;
	}

	memcpy(data + 0x10, test_buf + 4, 8);

	return 0;
}

static void make_png_key(unsigned char *key, const unsigned char *digest)
{
	int i;
	int value;

	for (i = 0; i < 4; i++) {
		value = (digest[i] >> 4) & 0xF;
		if (value < 10) {
			*key = '0' + value;
		} else {
			*key = 'a' + value - 10;
		}
		key++;

		value = digest[i] & 0xF;
		if (value < 10) {
			*key = '0' + value;
		} else {
			*key = 'a' + value - 10;
		}
		key++;
	}
}

/* EOF */
