#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "utils.h"
#ifdef WIN32
#include "my_printf.h"
#define	printf	my_printf
#endif


static const struct escape_table_t {
	char ch;
	char *escape;
	int len;
} escape_table[] = {
	{ '<', "&lt;", 4 },
	{ '>', "&gt;", 4 },
	{ '&', "&amp;", 5 },
	{ '\'', "&apos;", 6 },
	{ '\"', "&quot;", 6 }
};

static const char *escape_chars = "<>&'\"";


const char *get_escape_str(int ch)
{
	int i;

	for (i = 0; i < sizeof(escape_table) / sizeof(escape_table[0]); i++) {
		if (escape_table[i].ch == ch) {
			return escape_table[i].escape;
		}
	}

	return NULL;
}

int check_escape_chars(const char *text)
{
	if (strpbrk(text, escape_chars) != NULL) {
		return 1;
	}

	return 0;
}

char *escape_text(const char *text)
{
	long len = strlen(text);
	const char *ptr = text;
	char *new_text = NULL;
	const char *src;
	char *des;

	while ((ptr = strpbrk(ptr, escape_chars)) != NULL) {
		len += strlen(get_escape_str(*ptr));
		ptr++;
	}

	if ((new_text = (char *)malloc(len + 1)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", len + 1);
		return NULL;
	}

	src = text;
	des = new_text;

	while (*src) {
		if ((ptr = strpbrk(src, escape_chars)) == NULL) {
			while (*src) {
				*des++ = *src++;
			}
			break;
		}

		while (src < ptr) {
			*des++ = *src++;
		}

		ptr = get_escape_str(*src);
		src++;
		while (*ptr) {
			*des++ = *ptr++;
		}
	}

	*des = 0;

	return new_text;
}

int unescape_text(char *text, char *endptr)
{
	int result = -1;
	int len;
	char *des = text;
	unsigned long code;

	for ( ; *text != 0 && (endptr == NULL || text < endptr); text += len) {
		if ((len = get_next_utf8_char(&code, text, 1)) < 0) {
			printf("ERROR: 不正な文字です: %s\n", text);
			goto ERROR_EXIT;
		}
		if (code == 0) {
			break;
		}

		if (code <= 0x7F) {
			*des++ = (char)(code & 0x7F);
		} else if (des < text) {
			for ( ; 0 < len; len--) {
				*des++ = *text++;
			}
		} else {
			des += len;
		}
	}

	result = 0;

ERROR_EXIT:
	*des = 0;

	return result;
}

int get_next_utf8_char(unsigned long *code, const char *src, int check_escape)
{
	int len;
	int l;
	const char *save = src;
	int is_illegal = 0;

	*code = 0;

	if (src == NULL) {
		return 0;
	}

	if (*src == '&' && check_escape) {
		if (*(src + 1) == '#') {
			char *endptr;
			const char *start;
			if (*(src + 2) == 'x' || *(src + 2) == 'X') {
				start = src + 3;
				*code = strtoul(start, &endptr, 16);
			} else {
				start = src + 2;
				*code = strtoul(start, &endptr, 10);
			}
			if (endptr == NULL || *endptr != ';' || start == endptr) {
				printf("ERROR: 不正な文字参照です: %s\n", src);
				return -1;
			}
			endptr++;
			return endptr - src;
		} else {
			int i;
			for (i = 0; i < sizeof(escape_table) / sizeof(escape_table[0]); i++) {
				if (strncmp(src, escape_table[i].escape, escape_table[i].len) == 0) {
					*code = escape_table[i].ch;
					return escape_table[i].len;
				}
			}
			printf("ERROR: 未対応の実体参照です: %s\n", src);
			return -1;
		}
	}

	/* 1文字分をunsigned longにパック */
	*code = (unsigned char)*src;

	if (*src == '\0') {
		return 0;
	} else if ((*src & 0x80) == 0) {
		len = 1;
	} else if ((*src & 0x40) == 0) {
		/* illegal UTF-8 sequence */
		printf("ERROR: 不正な文字コードです: %02X\n", (int)*(unsigned char *)src);
		return -1;
	} else if ((*src & 0x20) == 0) {
		len = 2;
		*code &= 0x1F;
	} else if ((*src & 0x10) == 0) {
		len = 3;
		*code &= 0x0F;
	} else if ((*src & 0x08) == 0) {
		len = 4;
		*code &= 0x07;
	} else if ((*src & 0x04) == 0) {
		len = 5;
		*code &= 0x03;
	} else if ((*src & 0x02) == 0) {
		len = 6;
		*code &= 0x01;
	} else {
		/* illegal UTF-8 sequence */
		printf("ERROR: 不正な文字コードです: %02X\n", (int)*(unsigned char *)src);
		return -1;
	}
	src++;

	l = len - 1;
	while (0 < l--) {
		*code <<= 6;
		if ((*src & 0xC0) != 0x80) {
			/* illegal UTF-8 sequence */
			is_illegal = 1;
			break;
		}
		*code |= ((unsigned char)*src++) & 0x3F;
	}

	if (is_illegal ||
		(0xD800 <= *code && *code <= 0xDFFF) ||
		(len == 2 && *code <= 0x0000007F) ||
		(len == 3 && *code <= 0x000007FF) ||
		(len == 4 && *code <= 0x0000FFFF) ||
		(len == 5 && *code <= 0x001FFFFF) ||
		(len == 6 && *code <= 0x03FFFFFF)) {
		/* illegal UTF-8 sequence */
		printf("ERROR: 不正な文字コードです:");
		while (0 < len--) {
			printf(" %02X", (int)*(unsigned char *)save);
			if (*save == '\0') {
				break;
			}
			save++;
		}
		printf("\n");
		return -1;
	}

	return len;
}

char *skip_bom(char *src)
{
	const char bom[3] = { 0xEF, 0xBB, 0xBF };

	if (src[0] == bom[0] && src[1] == bom[1] && src[2] == bom[2]) {
		src += 3;
	}

	return src;
}

char *skip_space(char *src)
{
	while (*src == ' ' || *src == '\t') {
		src++;
	}

	return src;
}

char *skip_crlf(char *src)
{
	while (*src == '\r' || *src == '\n') {
		src++;
	}

	return src;
}

void trim_right(char *src)
{
	char *left = src;
	char *right = src + strlen(src);

	while (left < right) {
		right--;
		if (*right == ' ' || *right == '\t' || *right == '\r' || *right == '\n') {
			*right = 0;
		} else {
			break;
		}
	}
}

char *code_to_string(char *des, int *size, unsigned long code)
{
	int len;

	if (*size <= 0) {
		printf("BUG: code_to_string: buffer size=%d\n", *size);
		return NULL;
	}

	if (0x0000D800 <= code && code <= 0x0000DFFF) {
		printf("ERROR: 不正な文字コードです: %08lX\n", code);
		return NULL;
	} else if (code < 0x0000007F) {
		len = 1;
		*des++ = (char)(code & 0x7F);
	} else if (code < 0x000007FF) {
		len = 2;
		*des++ = (char)(0xC0 | (code >> 6));
	} else if (code < 0x0000FFFF) {
		len = 3;
		*des++ = (char)(0xE0 | (code >> 12));
	} else if (code < 0x001FFFFF) {
		len = 4;
		*des++ = (char)(0xF0 | (code >> 18));
	} else if (code < 0x03FFFFFF) {
		len = 5;
		*des++ = (char)(0xF8 | (code >> 24));
	} else if (code < 0x7FFFFFFF) {
		len = 6;
		*des++ = (char)(0xFC | (code >> 30));
	} else {
		printf("ERROR: 不正な文字コードです: %08lX\n", code);
		return NULL;
	}

	if (*size <= len) {
		printf("BUG: code_to_string: buffer size=%d, string length=%d\n", *size, len);
		return NULL;
	}
	*size -= len;

	len--;
	while (0 < len--) {
		*des++ = (char)(0x80 | ((code >> (6 * len)) & 0x3F));
	}
	*des = 0;

	return des;
}

/* EOF */
